from fastapi_users import FastAPIUsers

from fief.dependencies.users import authentication_backend, get_user_manager

fastapi_users = FastAPIUsers(
    get_user_manager,
    [authentication_backend],
)

current_active_user = fastapi_users.current_user(active=True)



from typing import Protocol

from pydantic import UUID4

from fief.models import M
from fief.schemas.generics import PM
from fief.services.webhooks.models import WebhookEvent, WebhookEventType
from fief.tasks import SendTask
from fief.tasks import trigger_webhooks as trigger_webhooks_task


def trigger_webhooks(
    event_type: type[WebhookEventType],
    object: M,
    schema_class: type[PM],
    *,
    workspace_id: UUID4,
    send_task: SendTask,
) -> None:
    event: WebhookEvent = WebhookEvent(
        type=event_type.key(), data=schema_class.from_orm(object).dict()
    )
    send_task(trigger_webhooks_task, workspace_id=str(workspace_id), event=event.json())


class TriggerWebhooks(Protocol):
    def __call__(
        self, event_type: type[WebhookEventType], object: M, schema_class: type[PM]
    ) -> None:
        ...



from fastapi import APIRouter, Depends, Header, Query, Request, status

from fief.apps.dashboard.dependencies import (
    BaseContext,
    DatatableColumn,
    DatatableQueryParameters,
    DatatableQueryParametersGetter,
    get_base_context,
)
from fief.apps.dashboard.forms.theme import (
    ThemeCreateForm,
    ThemePagePreviewForm,
    ThemeUpdateForm,
)
from fief.apps.dashboard.responses import HXRedirectResponse
from fief.dependencies.admin_authentication import is_authenticated_admin_session
from fief.dependencies.logger import get_audit_logger
from fief.dependencies.pagination import PaginatedObjects
from fief.dependencies.theme import (
    get_paginated_themes,
    get_theme_by_id_or_404,
    get_theme_preview,
)
from fief.dependencies.workspace_repositories import get_workspace_repository
from fief.forms import FormHelper
from fief.logger import AuditLogger
from fief.models import AuditLogMessage, Theme
from fief.repositories import TenantRepository, ThemeRepository
from fief.services.theme_preview import ThemePreview
from fief.templates import templates

router = APIRouter(dependencies=[Depends(is_authenticated_admin_session)])


async def get_columns() -> list[DatatableColumn]:
    return [
        DatatableColumn("Name", "name", "name_column", ordering="name"),
        DatatableColumn("Default", "default", "default_column"),
        DatatableColumn("Actions", "actions", "actions_column"),
    ]


async def get_list_context(
    columns: list[DatatableColumn] = Depends(get_columns),
    datatable_query_parameters: DatatableQueryParameters = Depends(
        DatatableQueryParametersGetter(["name", "default", "actions"])
    ),
    paginated_themes: PaginatedObjects[Theme] = Depends(get_paginated_themes),
):
    themes, count = paginated_themes
    return {
        "themes": themes,
        "count": count,
        "datatable_query_parameters": datatable_query_parameters,
        "columns": columns,
    }


async def get_list_template(hx_combobox: bool = Header(False)) -> str:
    if hx_combobox:
        return "admin/themes/list_combobox.html"
    return "admin/themes/list.html"


@router.get("/", name="dashboard.themes:list")
async def list_themes(
    template: str = Depends(get_list_template),
    list_context=Depends(get_list_context),
    context: BaseContext = Depends(get_base_context),
):
    return templates.TemplateResponse(template, {**context, **list_context})


@router.api_route("/create", methods=["GET", "POST"], name="dashboard.themes:create")
async def create_theme(
    request: Request,
    repository: ThemeRepository = Depends(get_workspace_repository(ThemeRepository)),
    list_context=Depends(get_list_context),
    context: BaseContext = Depends(get_base_context),
    audit_logger: AuditLogger = Depends(get_audit_logger),
):
    form_helper = FormHelper(
        ThemeCreateForm,
        "admin/themes/create.html",
        request=request,
        context={**context, **list_context},
    )

    if await form_helper.is_submitted_and_valid():
        form = await form_helper.get_form()

        theme = Theme.build_default()
        theme.default = False
        form.populate_obj(theme)

        theme = await repository.create(theme)
        audit_logger.log_object_write(AuditLogMessage.OBJECT_CREATED, theme)

        return HXRedirectResponse(
            request.url_for("dashboard.themes:update", id=theme.id),
            status_code=status.HTTP_201_CREATED,
            headers={"X-Fief-Object-Id": str(theme.id)},
        )

    return await form_helper.get_response()


@router.api_route(
    "/{id:uuid}/edit",
    methods=["GET", "POST"],
    name="dashboard.themes:update",
)
async def update_theme(
    request: Request,
    preview: str | None = Query(None),
    theme: Theme = Depends(get_theme_by_id_or_404),
    theme_preview: ThemePreview = Depends(get_theme_preview),
    repository: ThemeRepository = Depends(get_workspace_repository(ThemeRepository)),
    tenant_repository: TenantRepository = Depends(
        get_workspace_repository(TenantRepository)
    ),
    list_context=Depends(get_list_context),
    context: BaseContext = Depends(get_base_context),
    audit_logger: AuditLogger = Depends(get_audit_logger),
):
    form_helper = FormHelper(
        ThemeUpdateForm,
        "admin/themes/edit.html",
        object=theme,
        request=request,
        context={**context, **list_context, "theme": theme},
    )

    if await form_helper.is_submitted_and_valid():
        form = await form_helper.get_form()
        form.populate_obj(theme)

        if preview is None:
            await repository.update(theme)
            audit_logger.log_object_write(AuditLogMessage.OBJECT_UPDATED, theme)
            return HXRedirectResponse(request.url_for("dashboard.themes:list"))

    preview_page = preview if preview is not None else "login"
    tenant = await tenant_repository.get_default()
    assert tenant is not None
    form_helper.context["preview_content"] = await theme_preview.preview(
        preview_page, theme, tenant=tenant, request=request
    )

    page_preview_form = ThemePagePreviewForm(data={"page": preview_page})
    form_helper.context["page_preview_form"] = page_preview_form

    return await form_helper.get_response()


@router.post(
    "/{id:uuid}/default",
    name="dashboard.themes:default",
)
async def set_default_theme(
    request: Request,
    theme: Theme = Depends(get_theme_by_id_or_404),
    repository: ThemeRepository = Depends(get_workspace_repository(ThemeRepository)),
    audit_logger: AuditLogger = Depends(get_audit_logger),
):
    default_theme = await repository.get_default()
    if default_theme is not None and default_theme.id != theme.id:
        default_theme.default = False
        await repository.update(default_theme)
        audit_logger.log_object_write(AuditLogMessage.OBJECT_UPDATED, default_theme)

        theme.default = True
        await repository.update(theme)
        audit_logger.log_object_write(AuditLogMessage.OBJECT_UPDATED, theme)

    return HXRedirectResponse(request.url_for("dashboard.themes:list"))



from pydantic import UUID4
from sqlalchemy import ForeignKey
from sqlalchemy.orm import Mapped, mapped_column, relationship

from fief.models.base import MainBase
from fief.models.generics import GUID, CreatedUpdatedAt, UUIDModel
from fief.models.workspace import Workspace


class WorkspaceUser(UUIDModel, CreatedUpdatedAt, MainBase):
    __tablename__ = "workspace_users"

    workspace_id: Mapped[UUID4] = mapped_column(
        GUID, ForeignKey(Workspace.id, ondelete="CASCADE"), nullable=False
    )
    user_id: Mapped[UUID4] = mapped_column(GUID, nullable=False)

    workspace: Mapped[Workspace] = relationship(
        "Workspace", back_populates="workspace_users", lazy="joined"
    )

    def __repr__(self) -> str:
        return f"WorkspaceUser(id={self.id}, workspace_id={self.workspace_id}, user_id={self.user_id})"



from fief.repositories.admin_api_key import AdminAPIKeyRepository
from fief.repositories.admin_session_token import AdminSessionTokenRepository
from fief.repositories.authorization_code import AuthorizationCodeRepository
from fief.repositories.base import get_repository
from fief.repositories.client import ClientRepository
from fief.repositories.email_template import EmailTemplateRepository
from fief.repositories.grant import GrantRepository
from fief.repositories.login_session import LoginSessionRepository
from fief.repositories.oauth_account import OAuthAccountRepository
from fief.repositories.oauth_provider import OAuthProviderRepository
from fief.repositories.oauth_session import OAuthSessionRepository
from fief.repositories.permission import PermissionRepository
from fief.repositories.refresh_token import RefreshTokenRepository
from fief.repositories.registration_session import RegistrationSessionRepository
from fief.repositories.role import RoleRepository
from fief.repositories.session_token import SessionTokenRepository
from fief.repositories.tenant import TenantRepository
from fief.repositories.theme import ThemeRepository
from fief.repositories.user import UserRepository
from fief.repositories.user_field import UserFieldRepository
from fief.repositories.user_permission import UserPermissionRepository
from fief.repositories.user_role import UserRoleRepository
from fief.repositories.webhook import WebhookRepository
from fief.repositories.webhook_log import WebhookLogRepository
from fief.repositories.workspace import WorkspaceRepository
from fief.repositories.workspace_user import WorkspaceUserRepository

__all__ = [
    "WorkspaceRepository",
    "WorkspaceUserRepository",
    "AdminAPIKeyRepository",
    "AdminSessionTokenRepository",
    "AuthorizationCodeRepository",
    "ClientRepository",
    "EmailTemplateRepository",
    "GrantRepository",
    "LoginSessionRepository",
    "OAuthAccountRepository",
    "OAuthProviderRepository",
    "OAuthSessionRepository",
    "PermissionRepository",
    "RefreshTokenRepository",
    "RegistrationSessionRepository",
    "RoleRepository",
    "SessionTokenRepository",
    "TenantRepository",
    "ThemeRepository",
    "UserRepository",
    "UserFieldRepository",
    "UserPermissionRepository",
    "UserRoleRepository",
    "WebhookRepository",
    "WebhookLogRepository",
    "get_repository",
]



"""Add tokens lifetime parameter to Client

Revision ID: f50cd0e11d49
Revises: 31f6171a84fa
Create Date: 2023-01-16 13:09:39.792996

"""
import sqlalchemy as sa
from alembic import op

import fief
from fief.settings import settings

# revision identifiers, used by Alembic.
revision = "f50cd0e11d49"
down_revision = "31f6171a84fa"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column(
        "fief_clients",
        sa.Column(
            "authorization_code_lifetime_seconds",
            sa.Integer(),
            nullable=False,
            server_default=str(settings.default_authorization_code_lifetime_seconds),
        ),
    )
    op.add_column(
        "fief_clients",
        sa.Column(
            "access_id_token_lifetime_seconds",
            sa.Integer(),
            nullable=False,
            server_default=str(settings.default_access_id_token_lifetime_seconds),
        ),
    )
    op.add_column(
        "fief_clients",
        sa.Column(
            "refresh_token_lifetime_seconds",
            sa.Integer(),
            nullable=False,
            server_default=str(settings.default_refresh_token_lifetime_seconds),
        ),
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column("fief_clients", "refresh_token_lifetime_seconds")
    op.drop_column("fief_clients", "access_id_token_lifetime_seconds")
    op.drop_column("fief_clients", "authorization_code_lifetime_seconds")
    # ### end Alembic commands ###



from typing import Any

from pydantic import BaseModel

from fief.db import AsyncSession
from fief.repositories import TenantRepository, UserRepository
from fief.schemas.tenant import Tenant
from fief.schemas.user import UserEmailContext
from fief.services.email_template.types import EmailTemplateType


class EmailContext(BaseModel):
    tenant: Tenant
    user: UserEmailContext

    class Config:
        orm_mode = True

    @classmethod
    async def create_sample_context(cls, session: AsyncSession):
        context_kwargs = await cls._get_sample_context_kwargs(session)
        return cls(**context_kwargs)

    @classmethod
    async def _get_sample_context_kwargs(cls, session: AsyncSession) -> dict[str, Any]:
        context_kwargs: dict[str, Any] = {}
        tenant_repository = TenantRepository(session)
        tenant = await tenant_repository.get_default()
        assert tenant is not None
        context_kwargs["tenant"] = tenant

        user_repository = UserRepository(session)
        user = await user_repository.get_one_by_tenant(tenant.id)
        if user is None:
            context_kwargs["user"] = UserEmailContext.create_sample(tenant)
        else:
            context_kwargs["user"] = user

        return context_kwargs


class WelcomeContext(EmailContext):
    pass


class ForgotPasswordContext(EmailContext):
    reset_url: str

    @classmethod
    async def _get_sample_context_kwargs(cls, session: AsyncSession) -> dict[str, Any]:
        context_kwargs = await super()._get_sample_context_kwargs(session)
        context_kwargs["reset_url"] = "https://example.fief.dev/reset"
        return context_kwargs


EMAIL_TEMPLATE_CONTEXT_CLASS_MAP: dict[EmailTemplateType, type[EmailContext]] = {
    EmailTemplateType.BASE: EmailContext,
    EmailTemplateType.WELCOME: WelcomeContext,
    EmailTemplateType.FORGOT_PASSWORD: ForgotPasswordContext,
}



from fastapi import APIRouter, Depends, HTTPException, Request, status
from fastapi.responses import JSONResponse
from fastapi_users.exceptions import InvalidPasswordException, UserAlreadyExists

from fief.dependencies.user_field import get_update_user_fields
from fief.dependencies.users import UserManager, get_user_manager, get_user_update
from fief.errors import APIErrorCode
from fief.fastapi_users import current_active_user
from fief.models import User
from fief.models.user_field import UserField
from fief.schemas.user import UF, UserUpdate

router = APIRouter()


@router.get("/userinfo", name="user:userinfo.get")
async def get_userinfo(user: User = Depends(current_active_user)):
    return user.get_claims()


# OpenID specification requires the /userinfo endpoint
# to be available both with GET and POST methods 🤷‍♂️
# https://openid.net/specs/openid-connect-core-1_0.html#UserInfoRequest
@router.post("/userinfo", name="user:userinfo.post")
async def post_userinfo(user: User = Depends(current_active_user)):
    return user.get_claims()


@router.patch(
    "/profile",
    name="user:profile.patch",
    openapi_extra={
        "requestBody": {
            "content": {
                "application/json": {
                    "schema": UserUpdate.schema(
                        ref_template="#/paths/~1api~1profile/patch/requestBody/content/application~1json/schema/definitions/{model}"
                    )
                }
            },
            "required": True,
        }
    },
)
async def update_profile(
    request: Request,
    user_update: UserUpdate[UF] = Depends(get_user_update),
    update_user_fields: list[UserField] = Depends(get_update_user_fields),
    user: User = Depends(current_active_user),
    user_manager: UserManager = Depends(get_user_manager),
):
    try:
        user = await user_manager.update_with_fields(
            user_update,
            user,
            user_fields=update_user_fields,
            safe=True,
            request=request,
        )
    except InvalidPasswordException as e:
        # Build a JSON response manually to fine-tune the response structure
        return JSONResponse(
            content={
                "detail": APIErrorCode.USER_UPDATE_INVALID_PASSWORD,
                "reason": str(e.reason),
            },
            status_code=status.HTTP_400_BAD_REQUEST,
        )
    except UserAlreadyExists as e:
        raise HTTPException(
            status.HTTP_400_BAD_REQUEST,
            detail=APIErrorCode.USER_UPDATE_EMAIL_ALREADY_EXISTS,
        ) from e
    else:
        return user.get_claims()



from enum import StrEnum

from pydantic import BaseModel, HttpUrl

from fief.schemas.generics import CreatedUpdatedAt, UUIDSchema
from fief.services.webhooks.models import WEBHOOK_EVENTS

WebhookEventType = StrEnum("WebhookEventType", [event.key() for event in WEBHOOK_EVENTS])  # type: ignore


class WebhookCreate(BaseModel):
    url: HttpUrl
    events: list[WebhookEventType]


class WebhookUpdate(BaseModel):
    url: HttpUrl | None
    events: list[WebhookEventType] | None


class BaseWebhook(UUIDSchema, CreatedUpdatedAt):
    url: HttpUrl
    events: list[WebhookEventType]


class Webhook(BaseWebhook):
    pass


class WebhookSecret(BaseWebhook):
    secret: str



from pydantic import UUID4

from fief.schemas.generics import BaseModel, CreatedUpdatedAt
from fief.schemas.role import RoleEmbedded


class UserRoleCreate(BaseModel):
    id: UUID4


class BaseUserRole(CreatedUpdatedAt):
    user_id: UUID4
    role_id: UUID4


class UserRole(BaseUserRole):
    role: RoleEmbedded



"""Rename roles_permissions to fief_roles_permissions

Revision ID: 46fa2e6c46f7
Revises: 7f4cbde3bcb7
Create Date: 2022-06-08 15:25:26.573302

"""
import sqlalchemy as sa
from alembic import op
from sqlalchemy.dialects import postgresql

import fief

# revision identifiers, used by Alembic.
revision = "46fa2e6c46f7"
down_revision = "7f4cbde3bcb7"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.rename_table("roles_permissions", "fief_roles_permissions")
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.rename_table("fief_roles_permissions", "roles_permissions")
    # ### end Alembic commands ###



from fastapi import APIRouter, Depends, Header, Request, status

from fief import schemas
from fief.apps.dashboard.dependencies import (
    BaseContext,
    DatatableColumn,
    DatatableQueryParameters,
    DatatableQueryParametersGetter,
    get_base_context,
)
from fief.apps.dashboard.forms.user_field import (
    UserFieldConfigurationBase,
    UserFieldCreateForm,
    UserFieldUpdateForm,
)
from fief.apps.dashboard.responses import HXRedirectResponse
from fief.dependencies.admin_authentication import is_authenticated_admin_session
from fief.dependencies.logger import get_audit_logger
from fief.dependencies.pagination import PaginatedObjects
from fief.dependencies.user_field import (
    get_paginated_user_fields,
    get_user_field_by_id_or_404,
)
from fief.dependencies.webhooks import TriggerWebhooks, get_trigger_webhooks
from fief.dependencies.workspace_repositories import get_workspace_repository
from fief.forms import FormHelper
from fief.logger import AuditLogger
from fief.models import AuditLogMessage, OAuthProvider, UserField
from fief.repositories import UserFieldRepository
from fief.services.webhooks.models import (
    UserFieldCreated,
    UserFieldDeleted,
    UserFieldUpdated,
)
from fief.templates import templates

router = APIRouter(dependencies=[Depends(is_authenticated_admin_session)])


async def get_columns() -> list[DatatableColumn]:
    return [
        DatatableColumn("Name", "name", "name_column", ordering="name"),
        DatatableColumn("Slug", "slug", "slug_column", ordering="slug"),
        DatatableColumn("Type", "type", "type_column", ordering="type"),
    ]


async def get_list_context(
    columns: list[DatatableColumn] = Depends(get_columns),
    datatable_query_parameters: DatatableQueryParameters = Depends(
        DatatableQueryParametersGetter(["name", "slug", "type"])
    ),
    paginated_user_fields: PaginatedObjects[OAuthProvider] = Depends(
        get_paginated_user_fields
    ),
):
    user_fields, count = paginated_user_fields
    return {
        "user_fields": user_fields,
        "count": count,
        "datatable_query_parameters": datatable_query_parameters,
        "columns": columns,
    }


@router.get("/", name="dashboard.user_fields:list")
async def list_user_fields(
    list_context=Depends(get_list_context),
    context: BaseContext = Depends(get_base_context),
):
    return templates.TemplateResponse(
        "admin/user_fields/list.html", {**context, **list_context}
    )


@router.get("/{id:uuid}", name="dashboard.user_fields:get")
async def get_user_field(
    user_field: UserField = Depends(get_user_field_by_id_or_404),
    list_context=Depends(get_list_context),
    context: BaseContext = Depends(get_base_context),
):
    return templates.TemplateResponse(
        "admin/user_fields/get.html",
        {**context, **list_context, "user_field": user_field},
    )


@router.api_route(
    "/create", methods=["GET", "POST"], name="dashboard.user_fields:create"
)
async def create_user_field(
    request: Request,
    hx_trigger: str | None = Header(None),
    repository: UserFieldRepository = Depends(
        get_workspace_repository(UserFieldRepository)
    ),
    list_context=Depends(get_list_context),
    context: BaseContext = Depends(get_base_context),
    audit_logger: AuditLogger = Depends(get_audit_logger),
    trigger_webhooks: TriggerWebhooks = Depends(get_trigger_webhooks),
):
    form_class = await UserFieldCreateForm.get_form_class(request)
    form_helper = FormHelper(
        form_class,
        "admin/user_fields/create.html",
        request=request,
        context={**context, **list_context},
    )
    form = await form_helper.get_form()
    if hasattr(form, "configuration"):
        configuration_form: UserFieldConfigurationBase = form.configuration
        configuration_form.set_dynamic_parameters()

    if hx_trigger is None and await form_helper.is_submitted_and_valid():
        existing_user_field = await repository.get_by_slug(form.data["slug"])
        if existing_user_field is not None:
            form.slug.errors.append("A user field with this slug already exists.")
            return await form_helper.get_error_response(
                "A user field with this slug already exists.",
                "user_field_slug_already_exists",
            )

        user_field = UserField()
        form.populate_obj(user_field)
        user_field = await repository.create(user_field)
        audit_logger.log_object_write(AuditLogMessage.OBJECT_CREATED, user_field)
        trigger_webhooks(UserFieldCreated, user_field, schemas.user_field.UserField)

        return HXRedirectResponse(
            request.url_for("dashboard.user_fields:get", id=user_field.id),
            status_code=status.HTTP_201_CREATED,
            headers={"X-Fief-Object-Id": str(user_field.id)},
        )

    return await form_helper.get_response()


@router.api_route(
    "/{id:uuid}/edit",
    methods=["GET", "POST"],
    name="dashboard.user_fields:update",
)
async def update_user_field(
    request: Request,
    hx_trigger: str | None = Header(None),
    user_field: UserField = Depends(get_user_field_by_id_or_404),
    repository: UserFieldRepository = Depends(
        get_workspace_repository(UserFieldRepository)
    ),
    list_context=Depends(get_list_context),
    context: BaseContext = Depends(get_base_context),
    audit_logger: AuditLogger = Depends(get_audit_logger),
    trigger_webhooks: TriggerWebhooks = Depends(get_trigger_webhooks),
):
    form_class = await UserFieldUpdateForm.get_form_class(user_field)
    form_helper = FormHelper(
        form_class,
        "admin/user_fields/edit.html",
        object=user_field,
        request=request,
        context={**context, **list_context, "user_field": user_field},
    )
    form = await form_helper.get_form()
    configuration_form: UserFieldConfigurationBase = form.configuration
    configuration_form.set_dynamic_parameters()

    if hx_trigger is None and await form_helper.is_submitted_and_valid():
        existing_user_field = await repository.get_by_slug(form.data["slug"])
        if existing_user_field is not None and existing_user_field.id != user_field.id:
            form.slug.errors.append("A user field with this slug already exists.")
            return await form_helper.get_error_response(
                "A user field with this slug already exists.",
                "user_field_slug_already_exists",
            )

        form = await form_helper.get_form()
        form.populate_obj(user_field)

        await repository.update(user_field)
        audit_logger.log_object_write(AuditLogMessage.OBJECT_UPDATED, user_field)
        trigger_webhooks(UserFieldUpdated, user_field, schemas.user_field.UserField)

        return HXRedirectResponse(
            request.url_for("dashboard.user_fields:get", id=user_field.id)
        )

    return await form_helper.get_response()


@router.api_route(
    "/{id:uuid}/delete",
    methods=["GET", "DELETE"],
    name="dashboard.user_fields:delete",
)
async def delete_user_field(
    request: Request,
    user_field: UserField = Depends(get_user_field_by_id_or_404),
    repository: UserFieldRepository = Depends(
        get_workspace_repository(UserFieldRepository)
    ),
    list_context=Depends(get_list_context),
    context: BaseContext = Depends(get_base_context),
    audit_logger: AuditLogger = Depends(get_audit_logger),
    trigger_webhooks: TriggerWebhooks = Depends(get_trigger_webhooks),
):
    if request.method == "DELETE":
        await repository.delete(user_field)
        audit_logger.log_object_write(AuditLogMessage.OBJECT_DELETED, user_field)
        trigger_webhooks(UserFieldDeleted, user_field, schemas.user_field.UserField)

        return HXRedirectResponse(
            request.url_for("dashboard.user_fields:list"),
            status_code=status.HTTP_204_NO_CONTENT,
        )
    else:
        return templates.TemplateResponse(
            "admin/user_fields/delete.html",
            {**context, **list_context, "user_field": user_field},
        )



from fastapi import APIRouter, Depends, Header, Request, status

from fief import schemas
from fief.apps.dashboard.dependencies import (
    BaseContext,
    DatatableColumn,
    DatatableQueryParameters,
    DatatableQueryParametersGetter,
    get_base_context,
)
from fief.apps.dashboard.forms.oauth_provider import (
    OAuthProviderCreateForm,
    OAuthProviderUpdateForm,
)
from fief.apps.dashboard.responses import HXRedirectResponse
from fief.dependencies.admin_authentication import is_authenticated_admin_session
from fief.dependencies.logger import get_audit_logger
from fief.dependencies.oauth_provider import (
    get_oauth_provider_by_id_or_404,
    get_paginated_oauth_providers,
)
from fief.dependencies.pagination import PaginatedObjects
from fief.dependencies.webhooks import TriggerWebhooks, get_trigger_webhooks
from fief.dependencies.workspace_repositories import get_workspace_repository
from fief.forms import FormHelper
from fief.logger import AuditLogger
from fief.models import AuditLogMessage, OAuthProvider
from fief.repositories import OAuthProviderRepository
from fief.services.webhooks.models import (
    OAuthProviderCreated,
    OAuthProviderDeleted,
    OAuthProviderUpdated,
)
from fief.templates import templates

router = APIRouter(dependencies=[Depends(is_authenticated_admin_session)])


async def get_columns() -> list[DatatableColumn]:
    return [
        DatatableColumn("Provider", "provider", "provider_column", ordering="provider"),
        DatatableColumn("Name", "name", "name_column", ordering="name"),
    ]


async def get_list_context(
    columns: list[DatatableColumn] = Depends(get_columns),
    datatable_query_parameters: DatatableQueryParameters = Depends(
        DatatableQueryParametersGetter(["provider", "name"])
    ),
    paginated_oauth_providers: PaginatedObjects[OAuthProvider] = Depends(
        get_paginated_oauth_providers
    ),
):
    oauth_providers, count = paginated_oauth_providers
    return {
        "oauth_providers": oauth_providers,
        "count": count,
        "datatable_query_parameters": datatable_query_parameters,
        "columns": columns,
    }


async def get_list_template(hx_combobox: bool = Header(False)) -> str:
    if hx_combobox:
        return "admin/oauth_providers/list_combobox.html"
    return "admin/oauth_providers/list.html"


@router.get("/", name="dashboard.oauth_providers:list")
async def list_oauth_providers(
    template: str = Depends(get_list_template),
    list_context=Depends(get_list_context),
    context: BaseContext = Depends(get_base_context),
):
    return templates.TemplateResponse(template, {**context, **list_context})


@router.get("/{id:uuid}", name="dashboard.oauth_providers:get")
async def get_oauth_provider(
    oauth_provider: OAuthProvider = Depends(get_oauth_provider_by_id_or_404),
    list_context=Depends(get_list_context),
    context: BaseContext = Depends(get_base_context),
):
    return templates.TemplateResponse(
        "admin/oauth_providers/get.html",
        {**context, **list_context, "oauth_provider": oauth_provider},
    )


@router.api_route(
    "/create", methods=["GET", "POST"], name="dashboard.oauth_providers:create"
)
async def create_oauth_provider(
    request: Request,
    hx_trigger: str | None = Header(None),
    repository: OAuthProviderRepository = Depends(
        get_workspace_repository(OAuthProviderRepository)
    ),
    list_context=Depends(get_list_context),
    context: BaseContext = Depends(get_base_context),
    audit_logger: AuditLogger = Depends(get_audit_logger),
    trigger_webhooks: TriggerWebhooks = Depends(get_trigger_webhooks),
):
    form_helper = FormHelper(
        OAuthProviderCreateForm,
        "admin/oauth_providers/create.html",
        request=request,
        context={**context, **list_context},
    )

    if hx_trigger is None and await form_helper.is_submitted_and_valid():
        form = await form_helper.get_form()

        oauth_provider = OAuthProvider()
        form.populate_obj(oauth_provider)
        oauth_provider = await repository.create(oauth_provider)
        audit_logger.log_object_write(AuditLogMessage.OBJECT_CREATED, oauth_provider)
        trigger_webhooks(
            OAuthProviderCreated,
            oauth_provider,
            schemas.oauth_provider.OAuthProvider,
        )

        return HXRedirectResponse(
            request.url_for("dashboard.oauth_providers:get", id=oauth_provider.id),
            status_code=status.HTTP_201_CREATED,
            headers={"X-Fief-Object-Id": str(oauth_provider.id)},
        )

    return await form_helper.get_response()


@router.api_route(
    "/{id:uuid}/edit",
    methods=["GET", "POST"],
    name="dashboard.oauth_providers:update",
)
async def update_oauth_provider(
    request: Request,
    oauth_provider: OAuthProvider = Depends(get_oauth_provider_by_id_or_404),
    repository: OAuthProviderRepository = Depends(
        get_workspace_repository(OAuthProviderRepository)
    ),
    list_context=Depends(get_list_context),
    context: BaseContext = Depends(get_base_context),
    audit_logger: AuditLogger = Depends(get_audit_logger),
    trigger_webhooks: TriggerWebhooks = Depends(get_trigger_webhooks),
):
    form_helper = FormHelper(
        OAuthProviderUpdateForm,
        "admin/oauth_providers/edit.html",
        object=oauth_provider,
        request=request,
        context={**context, **list_context, "oauth_provider": oauth_provider},
    )

    if await form_helper.is_submitted_and_valid():
        form = await form_helper.get_form()
        form.populate_obj(oauth_provider)

        await repository.update(oauth_provider)
        audit_logger.log_object_write(AuditLogMessage.OBJECT_UPDATED, oauth_provider)
        trigger_webhooks(
            OAuthProviderUpdated,
            oauth_provider,
            schemas.oauth_provider.OAuthProvider,
        )

        return HXRedirectResponse(
            request.url_for("dashboard.oauth_providers:get", id=oauth_provider.id)
        )

    return await form_helper.get_response()


@router.api_route(
    "/{id:uuid}/delete",
    methods=["GET", "DELETE"],
    name="dashboard.oauth_providers:delete",
)
async def delete_oauth_provider(
    request: Request,
    oauth_provider: OAuthProvider = Depends(get_oauth_provider_by_id_or_404),
    repository: OAuthProviderRepository = Depends(
        get_workspace_repository(OAuthProviderRepository)
    ),
    list_context=Depends(get_list_context),
    context: BaseContext = Depends(get_base_context),
    audit_logger: AuditLogger = Depends(get_audit_logger),
    trigger_webhooks: TriggerWebhooks = Depends(get_trigger_webhooks),
):
    if request.method == "DELETE":
        await repository.delete(oauth_provider)
        audit_logger.log_object_write(AuditLogMessage.OBJECT_DELETED, oauth_provider)
        trigger_webhooks(
            OAuthProviderDeleted,
            oauth_provider,
            schemas.oauth_provider.OAuthProvider,
        )

        return HXRedirectResponse(
            request.url_for("dashboard.oauth_providers:list"),
            status_code=status.HTTP_204_NO_CONTENT,
        )
    else:
        return templates.TemplateResponse(
            "admin/oauth_providers/delete.html",
            {**context, **list_context, "oauth_provider": oauth_provider},
        )



from fastapi import APIRouter, Depends, HTTPException, Request, status
from fastapi_users.exceptions import InvalidPasswordException, UserAlreadyExists
from pydantic import UUID4

from fief import schemas, tasks
from fief.apps.dashboard.dependencies import (
    BaseContext,
    DatatableColumn,
    DatatableQueryParameters,
    DatatableQueryParametersGetter,
    get_base_context,
)
from fief.apps.dashboard.forms.user import (
    CreateUserPermissionForm,
    CreateUserRoleForm,
    UserAccessTokenForm,
    UserCreateForm,
    UserUpdateForm,
)
from fief.apps.dashboard.responses import HXRedirectResponse
from fief.crypto.access_token import generate_access_token
from fief.crypto.password import password_helper
from fief.db import AsyncSession
from fief.dependencies.admin_authentication import is_authenticated_admin_session
from fief.dependencies.current_workspace import (
    get_current_workspace,
    get_current_workspace_session,
)
from fief.dependencies.logger import get_audit_logger
from fief.dependencies.pagination import PaginatedObjects
from fief.dependencies.permission import (
    UserPermissionsGetter,
    get_user_permissions_getter,
)
from fief.dependencies.tasks import get_send_task
from fief.dependencies.user_field import get_user_fields
from fief.dependencies.users import (
    SQLAlchemyUserTenantDatabase,
    UserManager,
    get_admin_user_create_internal_model,
    get_admin_user_update_model,
    get_paginated_users,
    get_user_by_id_or_404,
    get_user_manager_from_user,
    get_user_oauth_accounts,
    get_user_permissions,
    get_user_roles,
)
from fief.dependencies.webhooks import TriggerWebhooks, get_trigger_webhooks
from fief.dependencies.workspace_repositories import get_workspace_repository
from fief.forms import FormHelper
from fief.logger import AuditLogger
from fief.models import (
    AuditLogMessage,
    OAuthAccount,
    User,
    UserField,
    UserPermission,
    UserRole,
    Workspace,
)
from fief.repositories import (
    ClientRepository,
    PermissionRepository,
    RoleRepository,
    TenantRepository,
    UserPermissionRepository,
    UserRepository,
    UserRoleRepository,
)
from fief.services.webhooks.models import (
    UserDeleted,
    UserPermissionCreated,
    UserPermissionDeleted,
    UserRoleCreated,
    UserRoleDeleted,
)
from fief.tasks import SendTask
from fief.templates import templates

router = APIRouter(dependencies=[Depends(is_authenticated_admin_session)])


async def get_columns(
    user_fields: list[UserField] = Depends(get_user_fields),
) -> list[DatatableColumn]:
    return [
        DatatableColumn("Email address", "email", "email_column", ordering="email"),
        DatatableColumn(
            "Created at", "created_at", "created_at_column", ordering="created_at"
        ),
        DatatableColumn(
            "Updated at", "updated_at", "updated_at_column", ordering="updated_at"
        ),
        DatatableColumn("ID", "id", "id_column", ordering="id"),
        DatatableColumn("Tenant", "tenant", "tenant_column", ordering="tenant.name"),
        *[
            DatatableColumn(
                user_field.name,
                user_field.slug,
                "user_field_value_column",
                renderer_macro_kwargs={"user_field": user_field},
            )
            for user_field in user_fields
        ],
    ]


async def get_list_context(
    columns: list[DatatableColumn] = Depends(get_columns),
    datatable_query_parameters: DatatableQueryParameters = Depends(
        DatatableQueryParametersGetter(["email", "created_at", "id", "tenant"])
    ),
    paginated_users: PaginatedObjects[User] = Depends(get_paginated_users),
):
    users, count = paginated_users

    return {
        "users": users,
        "count": count,
        "datatable_query_parameters": datatable_query_parameters,
        "columns": columns,
    }


@router.get("/", name="dashboard.users:list")
async def list_users(
    list_context=Depends(get_list_context),
    context: BaseContext = Depends(get_base_context),
):
    return templates.TemplateResponse(
        "admin/users/list.html", {**context, **list_context}
    )


@router.get("/{id:uuid}", name="dashboard.users:get")
async def get_user(
    user: User = Depends(get_user_by_id_or_404),
    list_context=Depends(get_list_context),
    context: BaseContext = Depends(get_base_context),
):
    return templates.TemplateResponse(
        "admin/users/get/account.html",
        {**context, **list_context, "user": user, "tab": "account"},
    )


@router.api_route("/create", methods=["GET", "POST"], name="dashboard.users:create")
async def create_user(
    request: Request,
    user_create_internal_model: type[
        schemas.user.UserCreateInternal[schemas.user.UF]
    ] = Depends(get_admin_user_create_internal_model),
    user_fields: list[UserField] = Depends(get_user_fields),
    tenant_repository: TenantRepository = Depends(
        get_workspace_repository(TenantRepository)
    ),
    list_context=Depends(get_list_context),
    context: BaseContext = Depends(get_base_context),
    audit_logger: AuditLogger = Depends(get_audit_logger),
    session: AsyncSession = Depends(get_current_workspace_session),
    workspace: Workspace = Depends(get_current_workspace),
    send_task: SendTask = Depends(get_send_task),
    trigger_webhooks: TriggerWebhooks = Depends(get_trigger_webhooks),
):
    form_class = await UserCreateForm.get_form_class(user_fields)
    form_helper = FormHelper(
        form_class,
        "admin/users/create.html",
        request=request,
        context={**context, **list_context},
    )

    if await form_helper.is_submitted_and_valid():
        form = await form_helper.get_form()

        tenant = await tenant_repository.get_by_id(form.data["tenant"])
        if tenant is None:
            form.tenant.errors.append("Unknown tenant.")
            return await form_helper.get_error_response(
                "Unknown tenant.", "unknown_tenant"
            )

        user_db = SQLAlchemyUserTenantDatabase(session, tenant, User)
        user_manager = UserManager(
            user_db,
            password_helper,
            workspace,
            tenant,
            send_task,
            audit_logger,
            trigger_webhooks,
        )
        user_create = user_create_internal_model(**form.data, tenant_id=tenant.id)

        try:
            user = await user_manager.create_with_fields(
                user_create, user_fields=user_fields, request=request
            )
            audit_logger.log_object_write(AuditLogMessage.OBJECT_CREATED, user)
        except UserAlreadyExists:
            form.email.errors.append(
                "A user with this email address already exists on this tenant."
            )
            return await form_helper.get_error_response(
                "A user with this email address already exists on this tenant.",
                "user_already_exists",
            )
        except InvalidPasswordException as e:
            form.password.errors.append(e.reason)
            return await form_helper.get_error_response(e.reason, "invalid_password")

        return HXRedirectResponse(
            request.url_for("dashboard.users:get", id=user.id),
            status_code=status.HTTP_201_CREATED,
            headers={"X-Fief-Object-Id": str(user.id)},
        )

    return await form_helper.get_response()


@router.api_route(
    "/{id:uuid}/edit",
    methods=["GET", "POST"],
    name="dashboard.users:update",
)
async def update_user(
    request: Request,
    user_update_model: type[schemas.user.UserUpdate[schemas.user.UF]] = Depends(
        get_admin_user_update_model
    ),
    user: User = Depends(get_user_by_id_or_404),
    user_fields: list[UserField] = Depends(get_user_fields),
    user_manager: UserManager = Depends(get_user_manager_from_user),
    list_context=Depends(get_list_context),
    context: BaseContext = Depends(get_base_context),
    audit_logger: AuditLogger = Depends(get_audit_logger),
    trigger_webhooks: TriggerWebhooks = Depends(get_trigger_webhooks),
):
    form_class = await UserUpdateForm.get_form_class(user_fields)
    form_helper = FormHelper(
        form_class,
        "admin/users/edit.html",
        object=user,
        request=request,
        context={**context, **list_context, "user": user},
    )

    if await form_helper.is_submitted_and_valid():
        form = await form_helper.get_form()
        data = form.data

        if data["password"] is None:
            data.pop("password")

        user_update = user_update_model(**data)

        try:
            user = await user_manager.update_with_fields(
                user_update, user, user_fields=user_fields, request=request
            )
            audit_logger.log_object_write(AuditLogMessage.OBJECT_UPDATED, user)
        except UserAlreadyExists:
            form.email.errors.append(
                "A user with this email address already exists on this tenant."
            )
            return await form_helper.get_error_response(
                "A user with this email address already exists on this tenant.",
                "user_already_exists",
            )
        except InvalidPasswordException as e:
            form.password.errors.append(e.reason)
            return await form_helper.get_error_response(e.reason, "invalid_password")

        return HXRedirectResponse(request.url_for("dashboard.users:get", id=user.id))

    return await form_helper.get_response()


@router.api_route(
    "/{id:uuid}/access-token",
    methods=["GET", "POST"],
    name="dashboard.users:access_token",
)
async def create_user_access_token(
    request: Request,
    user: User = Depends(get_user_by_id_or_404),
    get_user_permissions: UserPermissionsGetter = Depends(get_user_permissions_getter),
    client_repository: ClientRepository = Depends(
        get_workspace_repository(ClientRepository)
    ),
    workspace: Workspace = Depends(get_current_workspace),
    list_context=Depends(get_list_context),
    context: BaseContext = Depends(get_base_context),
    audit_logger: AuditLogger = Depends(get_audit_logger),
):
    form_helper = FormHelper(
        UserAccessTokenForm,
        "admin/users/access_token.html",
        request=request,
        context={**context, **list_context, "user": user},
    )
    form = await form_helper.get_form()
    form.client.query_endpoint_path = f"/admin/clients?tenant={user.tenant_id}"

    if await form_helper.is_submitted_and_valid():
        data = form.data
        tenant = user.tenant

        client = await client_repository.get_by_id(data["client"])
        if client is None or client.tenant_id != tenant.id:
            form.client.errors.append("Unknown client.")
            return await form_helper.get_error_response(
                "Unknown client.", "unknown_client"
            )
        form.client.data = client

        tenant_host = tenant.get_host(workspace.domain)
        permissions = await get_user_permissions(user)

        access_token = generate_access_token(
            user.tenant.get_sign_jwk(),
            tenant_host,
            client,
            user,
            data["scopes"],
            permissions,
            client.access_id_token_lifetime_seconds,
        )

        audit_logger(
            AuditLogMessage.USER_TOKEN_GENERATED_BY_ADMIN,
            subject_user_id=user.id,
            scope=data["scopes"],
        )

        return templates.TemplateResponse(
            "admin/users/access_token_result.html",
            {
                **context,
                **list_context,
                "user": user,
                "access_token": access_token,
                "expires_in": client.access_id_token_lifetime_seconds,
            },
        )

    return await form_helper.get_response()


@router.api_route(
    "/{id:uuid}/delete",
    methods=["GET", "DELETE"],
    name="dashboard.users:delete",
)
async def delete_user(
    request: Request,
    user: User = Depends(get_user_by_id_or_404),
    repository: UserRepository = Depends(get_workspace_repository(UserRepository)),
    list_context=Depends(get_list_context),
    context: BaseContext = Depends(get_base_context),
    audit_logger: AuditLogger = Depends(get_audit_logger),
    trigger_webhooks: TriggerWebhooks = Depends(get_trigger_webhooks),
):
    if request.method == "DELETE":
        await repository.delete(user)
        audit_logger.log_object_write(AuditLogMessage.OBJECT_DELETED, user)
        trigger_webhooks(UserDeleted, user, schemas.user.UserRead)

        return HXRedirectResponse(
            request.url_for("dashboard.users:list"),
            status_code=status.HTTP_204_NO_CONTENT,
        )
    else:
        return templates.TemplateResponse(
            "admin/users/delete.html", {**context, **list_context, "user": user}
        )


@router.api_route(
    "/{id:uuid}/permissions",
    methods=["GET", "POST"],
    name="dashboard.users:permissions",
)
async def user_permissions(
    request: Request,
    user: User = Depends(get_user_by_id_or_404),
    user_permissions: list[UserPermission] = Depends(get_user_permissions),
    list_context=Depends(get_list_context),
    context: BaseContext = Depends(get_base_context),
    permission_repository: PermissionRepository = Depends(
        get_workspace_repository(PermissionRepository)
    ),
    user_permission_repository: UserPermissionRepository = Depends(
        get_workspace_repository(UserPermissionRepository)
    ),
    audit_logger: AuditLogger = Depends(get_audit_logger),
    trigger_webhooks: TriggerWebhooks = Depends(get_trigger_webhooks),
):
    form_helper = FormHelper(
        CreateUserPermissionForm,
        "admin/users/get/permissions.html",
        request=request,
        context={
            **context,
            **list_context,
            "user": user,
            "user_permissions": user_permissions,
            "tab": "permissions",
        },
    )

    if await form_helper.is_submitted_and_valid():
        form = await form_helper.get_form()
        permission_id = form.data["permission"]

        permission = await permission_repository.get_by_id(permission_id)
        if permission is None:
            form.permission.errors.append("Unknown permission.")
            return await form_helper.get_error_response(
                "Unknown permission.", "unknown_permission"
            )

        existing_user_permission = (
            await user_permission_repository.get_by_permission_and_user(
                user.id, permission_id, direct_only=True
            )
        )
        if existing_user_permission is not None:
            form.permission.errors.append(
                "This permission is already granted to this user."
            )
            return await form_helper.get_error_response(
                "This permission is already granted to this user.",
                "already_added_permission",
            )

        user_permission = UserPermission(user_id=user.id, permission=permission)
        await user_permission_repository.create(user_permission)
        audit_logger.log_object_write(
            AuditLogMessage.OBJECT_CREATED,
            user_permission,
            subject_user_id=user.id,
            permission_id=str(permission.id),
        )
        trigger_webhooks(
            UserPermissionCreated,
            user_permission,
            schemas.user_permission.UserPermission,
        )

        return HXRedirectResponse(
            request.url_for("dashboard.users:permissions", id=user.id),
            status_code=status.HTTP_201_CREATED,
            headers={"X-Fief-Object-Id": str(user_permission.id)},
        )

    return await form_helper.get_response()


@router.delete(
    "/{id:uuid}/permissions/{permission_id:uuid}",
    name="dashboard.users:delete_permission",
)
async def delete_user_permission(
    request: Request,
    permission_id: UUID4,
    user: User = Depends(get_user_by_id_or_404),
    user_permission_repository: UserPermissionRepository = Depends(
        get_workspace_repository(UserPermissionRepository)
    ),
    audit_logger: AuditLogger = Depends(get_audit_logger),
    trigger_webhooks: TriggerWebhooks = Depends(get_trigger_webhooks),
):
    user_permission = await user_permission_repository.get_by_permission_and_user(
        user.id, permission_id, direct_only=True
    )
    if user_permission is None:
        raise HTTPException(status_code=status.HTTP_404_NOT_FOUND)

    await user_permission_repository.delete(user_permission)
    audit_logger.log_object_write(
        AuditLogMessage.OBJECT_DELETED,
        user_permission,
        subject_user_id=user.id,
        permission_id=str(permission_id),
    )
    trigger_webhooks(
        UserPermissionDeleted,
        user_permission,
        schemas.user_permission.UserPermission,
    )

    return HXRedirectResponse(
        request.url_for("dashboard.users:permissions", id=user.id),
        status_code=status.HTTP_204_NO_CONTENT,
    )


@router.api_route(
    "/{id:uuid}/roles", methods=["GET", "POST"], name="dashboard.users:roles"
)
async def user_roles(
    request: Request,
    user: User = Depends(get_user_by_id_or_404),
    user_roles: list[UserRole] = Depends(get_user_roles),
    list_context=Depends(get_list_context),
    context: BaseContext = Depends(get_base_context),
    role_repository: RoleRepository = Depends(get_workspace_repository(RoleRepository)),
    user_role_repository: UserRoleRepository = Depends(
        get_workspace_repository(UserRoleRepository)
    ),
    workspace: Workspace = Depends(get_current_workspace),
    send_task: SendTask = Depends(get_send_task),
    audit_logger: AuditLogger = Depends(get_audit_logger),
    trigger_webhooks: TriggerWebhooks = Depends(get_trigger_webhooks),
):
    form_helper = FormHelper(
        CreateUserRoleForm,
        "admin/users/get/roles.html",
        request=request,
        context={
            **context,
            **list_context,
            "user": user,
            "user_roles": user_roles,
            "tab": "roles",
        },
    )

    if await form_helper.is_submitted_and_valid():
        form = await form_helper.get_form()
        role_id = form.data["role"]

        role = await role_repository.get_by_id(role_id)
        if role is None:
            form.role.errors.append("Unknown role.")
            return await form_helper.get_error_response("Unknown role.", "unknown_role")

        existing_user_role = await user_role_repository.get_by_role_and_user(
            user.id, role_id
        )
        if existing_user_role is not None:
            form.role.errors.append("This role is already granted to this user.")
            return await form_helper.get_error_response(
                "This role is already granted to this user.", "already_added_role"
            )

        user_role = UserRole(user_id=user.id, role=role)
        await user_role_repository.create(user_role)
        audit_logger.log_object_write(
            AuditLogMessage.OBJECT_CREATED,
            user_role,
            subject_user_id=user.id,
            role_id=str(role.id),
        )
        trigger_webhooks(UserRoleCreated, user_role, schemas.user_role.UserRole)

        send_task(
            tasks.on_user_role_created, str(user.id), str(role.id), str(workspace.id)
        )

        return HXRedirectResponse(
            request.url_for("dashboard.users:roles", id=user.id),
            status_code=status.HTTP_201_CREATED,
            headers={"X-Fief-Object-Id": str(user_role.id)},
        )

    return await form_helper.get_response()


@router.delete("/{id:uuid}/roles/{role_id:uuid}", name="dashboard.users:delete_role")
async def delete_user_role(
    request: Request,
    role_id: UUID4,
    user: User = Depends(get_user_by_id_or_404),
    user_role_repository: UserRoleRepository = Depends(
        get_workspace_repository(UserRoleRepository)
    ),
    workspace: Workspace = Depends(get_current_workspace),
    send_task: SendTask = Depends(get_send_task),
    audit_logger: AuditLogger = Depends(get_audit_logger),
    trigger_webhooks: TriggerWebhooks = Depends(get_trigger_webhooks),
):
    user_role = await user_role_repository.get_by_role_and_user(user.id, role_id)
    if user_role is None:
        raise HTTPException(status_code=status.HTTP_404_NOT_FOUND)

    await user_role_repository.delete(user_role)
    audit_logger.log_object_write(
        AuditLogMessage.OBJECT_DELETED,
        user_role,
        subject_user_id=user.id,
        role_id=str(role_id),
    )
    trigger_webhooks(UserRoleDeleted, user_role, schemas.user_role.UserRole)

    send_task(tasks.on_user_role_deleted, str(user.id), str(role_id), str(workspace.id))

    return HXRedirectResponse(
        request.url_for("dashboard.users:roles", id=user.id),
        status_code=status.HTTP_204_NO_CONTENT,
    )


@router.get("/{id:uuid}/oauth-accounts", name="dashboard.users:oauth_accounts")
async def user_oauth_accounts(
    user: User = Depends(get_user_by_id_or_404),
    oauth_accounts: list[OAuthAccount] = Depends(get_user_oauth_accounts),
    list_context=Depends(get_list_context),
    context: BaseContext = Depends(get_base_context),
):
    return templates.TemplateResponse(
        "admin/users/get/oauth_accounts.html",
        {
            **context,
            **list_context,
            "user": user,
            "oauth_accounts": oauth_accounts,
            "tab": "oauth",
        },
    )



from fastapi import Depends

from fief.db.workspace import WorkspaceEngineManager
from fief.dependencies.db import get_workspace_engine_manager
from fief.dependencies.main_repositories import get_main_repository
from fief.dependencies.workspace_db import get_workspace_db
from fief.repositories import WorkspaceRepository, WorkspaceUserRepository
from fief.services.workspace_creation import WorkspaceCreation
from fief.services.workspace_db import WorkspaceDatabase


async def get_workspace_creation(
    workspace_repository: WorkspaceRepository = Depends(
        get_main_repository(WorkspaceRepository)
    ),
    workspace_user_repository: WorkspaceUserRepository = Depends(
        get_main_repository(WorkspaceUserRepository)
    ),
    workspace_db: WorkspaceDatabase = Depends(get_workspace_db),
    workspace_engine_manager: WorkspaceEngineManager = Depends(
        get_workspace_engine_manager
    ),
) -> WorkspaceCreation:
    return WorkspaceCreation(
        workspace_repository,
        workspace_user_repository,
        workspace_db,
        workspace_engine_manager,
    )



from datetime import datetime
from enum import StrEnum

from pydantic import UUID4
from sqlalchemy import JSON, String, Text
from sqlalchemy.orm import Mapped, mapped_column, relationship

from fief.models.base import WorkspaceBase
from fief.models.generics import GUID, TIMESTAMPAware, UUIDModel
from fief.models.user import User


class AuditLogMessage(StrEnum):
    OBJECT_CREATED = "OBJECT_CREATED"
    OBJECT_UPDATED = "OBJECT_UPDATED"
    OBJECT_DELETED = "OBJECT_DELETED"
    USER_REGISTERED = "USER_REGISTERED"
    USER_UPDATED = "USER_UPDATED"
    USER_FORGOT_PASSWORD_REQUESTED = "USER_FORGOT_PASSWORD_REQUESTED"
    USER_PASSWORD_RESET = "USER_PASSWORD_RESET"
    USER_TOKEN_GENERATED = "USER_TOKEN_GENERATED"
    USER_TOKEN_GENERATED_BY_ADMIN = "USER_TOKEN_GENERATED_BY_ADMIN"
    OAUTH_PROVIDER_USER_ACCESS_TOKEN_GET = "OAUTH_PROVIDER_USER_ACCESS_TOKEN_GET"


class AuditLog(UUIDModel, WorkspaceBase):
    __tablename__ = "audit_logs"

    timestamp: Mapped[datetime] = mapped_column(
        TIMESTAMPAware(timezone=True), nullable=False, index=True
    )
    level: Mapped[str] = mapped_column(String(length=255), nullable=False, index=True)
    message: Mapped[str] = mapped_column(Text, nullable=False)
    extra: Mapped[dict] = mapped_column(JSON, nullable=True)

    subject_user_id: Mapped[UUID4 | None] = mapped_column(
        GUID, nullable=True, index=True
    )

    object_id: Mapped[UUID4 | None] = mapped_column(GUID, nullable=True, index=True)
    object_class: Mapped[str | None] = mapped_column(
        String(length=255), nullable=True, index=True
    )

    admin_user_id: Mapped[UUID4 | None] = mapped_column(GUID, nullable=True, index=True)
    admin_api_key_id: Mapped[UUID4 | None] = mapped_column(
        GUID, nullable=True, index=True
    )

    subject_user: Mapped[User | None] = relationship(
        "User",
        # Define a relationship, but without a foreign key constraint
        foreign_keys="AuditLog.subject_user_id",
        primaryjoin="AuditLog.subject_user_id == User.id",
    )



from pydantic import UUID4
from sqlalchemy import select
from sqlalchemy.sql import Select

from fief.models import OAuthAccount
from fief.repositories.base import BaseRepository, UUIDRepositoryMixin


class OAuthAccountRepository(
    BaseRepository[OAuthAccount], UUIDRepositoryMixin[OAuthAccount]
):
    model = OAuthAccount

    async def get_by_provider_and_account_id(
        self, provider_id: UUID4, account_id: str
    ) -> OAuthAccount | None:
        statement = select(OAuthAccount).where(
            OAuthAccount.oauth_provider_id == provider_id,
            OAuthAccount.account_id == account_id,
        )
        return await self.get_one_or_none(statement)

    async def get_by_provider_and_user(
        self, provider_id: UUID4, user: UUID4
    ) -> OAuthAccount | None:
        statement = select(OAuthAccount).where(
            OAuthAccount.oauth_provider_id == provider_id,
            OAuthAccount.user_id == user,
        )
        return await self.get_one_or_none(statement)

    def get_by_user_statement(self, user: UUID4) -> Select:
        statement = select(OAuthAccount).where(OAuthAccount.user_id == user)
        return statement



from sqlalchemy import select

from fief.models import AdminSessionToken
from fief.repositories.base import BaseRepository, UUIDRepositoryMixin


class AdminSessionTokenRepository(
    BaseRepository[AdminSessionToken], UUIDRepositoryMixin[AdminSessionToken]
):
    model = AdminSessionToken

    async def get_by_token(self, token: str) -> AdminSessionToken | None:
        statement = select(AdminSessionToken).where(AdminSessionToken.token == token)
        return await self.get_one_or_none(statement)



from sqlalchemy import JSON, String, Text
from sqlalchemy.orm import Mapped, mapped_column

from fief.crypto.encryption import FernetEngine, StringEncryptedType
from fief.models.base import WorkspaceBase
from fief.models.generics import CreatedUpdatedAt, UUIDModel
from fief.services.oauth_provider import AvailableOAuthProvider
from fief.settings import settings


class OAuthProvider(UUIDModel, CreatedUpdatedAt, WorkspaceBase):
    __tablename__ = "oauth_providers"

    provider: Mapped[AvailableOAuthProvider] = mapped_column(
        String(length=255), nullable=False
    )

    client_id: Mapped[str] = mapped_column(
        StringEncryptedType(Text, settings.encryption_key, FernetEngine), nullable=False
    )
    client_secret: Mapped[str] = mapped_column(
        StringEncryptedType(Text, settings.encryption_key, FernetEngine), nullable=False
    )
    scopes: Mapped[list[str]] = mapped_column(JSON, nullable=False, default=list)
    name: Mapped[str | None] = mapped_column(String(length=255), nullable=True)
    openid_configuration_endpoint: Mapped[str | None] = mapped_column(
        Text, nullable=True
    )

    def get_provider_display_name(self) -> str:
        return AvailableOAuthProvider[self.provider].get_display_name()

    @property
    def display_name(self) -> str:
        return f"{self.get_provider_display_name()}{f' ({self.name})' if self.name else ''}"



"""Add registration_allowed column to Tenant

Revision ID: 67c0904f3ebc
Revises: 8d7729f5ed1f
Create Date: 2022-10-28 13:24:29.274106

"""
import sqlalchemy as sa
from alembic import op

import fief

# revision identifiers, used by Alembic.
revision = "67c0904f3ebc"
down_revision = "8d7729f5ed1f"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column(
        "fief_tenants",
        sa.Column(
            "registration_allowed", sa.Boolean(), nullable=False, server_default="1"
        ),
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column("fief_tenants", "registration_allowed")
    # ### end Alembic commands ###



from typing import TYPE_CHECKING, Any, Optional

from pydantic import UUID4
from sqlalchemy import Boolean, ForeignKey, String
from sqlalchemy.orm import Mapped, mapped_column, relationship
from sqlalchemy.sql.schema import UniqueConstraint

from fief.models.base import WorkspaceBase
from fief.models.generics import GUID, CreatedUpdatedAt, UUIDModel
from fief.models.tenant import Tenant
from fief.models.user_field import UserField

if TYPE_CHECKING:
    from fief.models.user_field_value import UserFieldValue


class User(UUIDModel, CreatedUpdatedAt, WorkspaceBase):
    __tablename__ = "users"
    __table_args__ = (UniqueConstraint("email", "tenant_id"),)

    if TYPE_CHECKING:
        email: str
        hashed_password: str
        is_active: bool
        is_superuser: bool
        is_verified: bool
    else:
        email: Mapped[str] = mapped_column(
            String(length=320), index=True, nullable=False
        )
        hashed_password: Mapped[str] = mapped_column(String(length=255), nullable=False)
        is_active: Mapped[bool] = mapped_column(Boolean, default=True, nullable=False)
        is_superuser: Mapped[bool] = mapped_column(
            Boolean, default=False, nullable=False
        )
        is_verified: Mapped[bool] = mapped_column(
            Boolean, default=False, nullable=False
        )

    tenant_id: Mapped[UUID4] = mapped_column(
        GUID, ForeignKey(Tenant.id, ondelete="CASCADE"), nullable=False
    )
    tenant: Mapped[Tenant] = relationship("Tenant")

    user_field_values: Mapped[list["UserFieldValue"]] = relationship(
        "UserFieldValue", back_populates="user", cascade="all, delete", lazy="selectin"
    )

    def __repr__(self) -> str:
        return f"User(id={self.id}, email={self.email})"

    @property
    def fields(self) -> dict[str, Any]:
        return dict(
            user_field_value.get_slug_and_value()
            for user_field_value in self.user_field_values
        )

    def get_user_field_value(self, user_field: UserField) -> Optional["UserFieldValue"]:
        for user_field_value in self.user_field_values:
            if user_field_value.user_field_id == user_field.id:
                return user_field_value
        return None

    def get_claims(self) -> dict[str, Any]:
        fields = dict(
            user_field_value.get_slug_and_value(json_serializable=True)
            for user_field_value in self.user_field_values
        )
        return {
            "sub": str(self.id),
            "email": self.email,
            "tenant_id": str(self.tenant_id),
            "is_active": self.is_active,
            "is_superuser": self.is_superuser,
            "is_verified": self.is_verified,
            "fields": fields,
        }



import functools

from starlette.datastructures import MutableHeaders
from starlette.types import ASGIApp, Message, Receive, Scope, Send


class SecurityHeadersMiddleware:
    def __init__(self, app: ASGIApp):
        self.app = app

    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
        if scope["type"] != "http":  # pragma: no cover
            await self.app(scope, receive, send)
            return

        send = functools.partial(self.send, send=send, scope=scope)
        await self.app(scope, receive, send)

    async def send(self, message: Message, send: Send, scope: Scope) -> None:
        if message["type"] != "http.response.start":
            await send(message)
            return

        message.setdefault("headers", [])
        headers = MutableHeaders(scope=message)

        headers.append("content-security-policy", "frame-ancestors 'self'")
        headers.append("x-frame-options", "SAMEORIGIN")
        headers.append("referrer-policy", "strict-origin-when-cross-origin")
        headers.append("x-content-type-options", "nosniff")
        headers.append("permissions-policy", "geolocation=(), camera=(), microphone=()")

        await send(message)



import secrets

from sqlalchemy import JSON, String
from sqlalchemy.orm import Mapped, mapped_column

from fief.models.base import WorkspaceBase
from fief.models.generics import CreatedUpdatedAt, UUIDModel


class Webhook(UUIDModel, CreatedUpdatedAt, WorkspaceBase):
    __tablename__ = "webhooks"

    url: Mapped[str] = mapped_column(String(length=255), nullable=False)
    secret: Mapped[str] = mapped_column(
        String(length=255), default=secrets.token_urlsafe, nullable=False
    )
    events: Mapped[list[str]] = mapped_column(JSON, nullable=False, default=list)

    def regenerate_secret(self) -> str:
        self.secret = secrets.token_urlsafe()
        return self.secret



from fastapi import APIRouter, Depends, Request, Response, status
from fastapi.responses import RedirectResponse
from fastapi_users.exceptions import InvalidPasswordException, UserAlreadyExists
from fastapi_users.router import ErrorCode

from fief.apps.auth.forms.register import RF, get_register_form_class
from fief.dependencies.auth import get_login_session
from fief.dependencies.authentication_flow import get_authentication_flow
from fief.dependencies.oauth_provider import get_oauth_providers
from fief.dependencies.register import (
    get_optional_registration_session,
    get_registration_flow,
)
from fief.dependencies.tenant import get_current_tenant
from fief.dependencies.theme import get_current_theme
from fief.exceptions import LoginException
from fief.forms import FormHelper
from fief.locale import gettext_lazy as _
from fief.models import (
    OAuthProvider,
    RegistrationSession,
    RegistrationSessionFlow,
    Tenant,
    Theme,
)
from fief.schemas.auth import LoginError
from fief.services.authentication_flow import AuthenticationFlow
from fief.services.registration_flow import RegistrationFlow

router = APIRouter()


@router.api_route(
    "/register",
    methods=["GET", "POST"],
    name="register:register",
    dependencies=[Depends(get_login_session)],
)
async def register(
    request: Request,
    register_form_class: type[RF] = Depends(get_register_form_class),
    registration_flow: RegistrationFlow = Depends(get_registration_flow),
    authentication_flow: AuthenticationFlow = Depends(get_authentication_flow),
    registration_session: RegistrationSession
    | None = Depends(get_optional_registration_session),
    oauth_providers: list[OAuthProvider] | None = Depends(get_oauth_providers),
    tenant: Tenant = Depends(get_current_tenant),
    theme: Theme = Depends(get_current_theme),
):
    if not tenant.registration_allowed:
        raise LoginException(
            LoginError.get_registration_disabled(_("Registration is disabled")),
            fatal=True,
        )

    response: Response
    form_helper = FormHelper(
        register_form_class,
        "auth/register.html",
        request=request,
        context={
            "finalize": registration_session is not None
            and registration_session.flow != RegistrationSessionFlow.PASSWORD,
            "oauth_providers": oauth_providers,
            "tenant": tenant,
            "theme": theme,
        },
    )
    form = await form_helper.get_form()

    if (
        request.method != "POST"
        and registration_session is not None
        and registration_session.email
    ):
        form.email.data = registration_session.email

    if registration_session is not None and await form_helper.is_submitted_and_valid():
        try:
            user = await registration_flow.create_user(
                form.data, tenant, registration_session, request=request
            )
        except UserAlreadyExists:
            return await form_helper.get_error_response(
                _("A user with the same email address already exists."),
                error_code=ErrorCode.REGISTER_USER_ALREADY_EXISTS,
            )
        except InvalidPasswordException as e:
            form.password.errors.append(e.reason)
            return await form_helper.get_error_response(e.reason, "invalid_password")
        else:
            response = RedirectResponse(
                tenant.url_path_for(request, "auth:consent"),
                status_code=status.HTTP_302_FOUND,
            )
            response = await authentication_flow.create_session_token(response, user.id)
            response = await registration_flow.delete_registration_session(
                response, registration_session
            )
            return response

    response = await form_helper.get_response()
    if registration_session is None:
        await registration_flow.create_registration_session(
            response, flow=RegistrationSessionFlow.PASSWORD, tenant=tenant
        )
    return response



from fastapi import Depends

from fief.dependencies.admin_session import get_admin_session_token
from fief.dependencies.main_repositories import get_main_repository
from fief.models import AdminSessionToken, Workspace
from fief.repositories import WorkspaceRepository


async def get_admin_user_workspaces(
    repository: WorkspaceRepository = Depends(get_main_repository(WorkspaceRepository)),
    admin_session_token: AdminSessionToken = Depends(get_admin_session_token),
) -> list[Workspace]:
    return await repository.get_by_admin_user(admin_session_token.user_id)



import uuid

import dramatiq
from dramatiq.middleware import CurrentMessage

from fief.models import Webhook
from fief.repositories import WebhookLogRepository, WebhookRepository
from fief.services.webhooks.delivery import WebhookDelivery, WebhookDeliveryError
from fief.services.webhooks.models import WebhookEvent
from fief.settings import settings
from fief.tasks.base import ObjectDoesNotExistTaskError, TaskBase


class DeliverWebhookTask(TaskBase):
    __name__ = "deliver_webhook"

    async def run(self, workspace_id: str, webhook_id: str, event: str):
        workspace = await self._get_workspace(uuid.UUID(workspace_id))
        async with self.get_workspace_session(workspace) as session:
            webhook_repository = WebhookRepository(session)
            webhook = await webhook_repository.get_by_id(uuid.UUID(webhook_id))

            if webhook is None:
                raise ObjectDoesNotExistTaskError(Webhook, webhook_id)

            message = CurrentMessage.get_current_message()
            retries = message.options.get("retries", 0)

            webhook_log_repository = WebhookLogRepository(session)
            webhook_delivery = WebhookDelivery(webhook_log_repository)
            parsed_event = WebhookEvent.parse_raw(event)
            await webhook_delivery.deliver(webhook, parsed_event, attempt=retries + 1)


def should_retry_deliver_webhook(retries_so_far, exception):
    return retries_so_far < settings.webhooks_max_attempts and isinstance(
        exception, WebhookDeliveryError
    )


deliver_webhook = dramatiq.actor(
    DeliverWebhookTask(), retry_when=should_retry_deliver_webhook
)


class TriggerWebhooksTask(TaskBase):
    __name__ = "trigger_webhooks"

    async def run(self, workspace_id: str, event: str):
        workspace = await self._get_workspace(uuid.UUID(workspace_id))
        async with self.get_workspace_session(workspace) as session:
            webhook_repository = WebhookRepository(session)
            webhooks = await webhook_repository.all()
            parsed_event = WebhookEvent.parse_raw(event)
            for webhook in webhooks:
                if parsed_event.type in webhook.events:
                    self.send_task(
                        deliver_webhook,
                        workspace_id=workspace_id,
                        webhook_id=str(webhook.id),
                        event=event,
                    )


trigger_webhooks = dramatiq.actor(TriggerWebhooksTask())



from fastapi import APIRouter, Depends, Query, Request, status
from fastapi.responses import RedirectResponse
from fastapi_users.exceptions import (
    InvalidPasswordException,
    InvalidResetPasswordToken,
    UserInactive,
    UserNotExists,
)

from fief.apps.auth.forms.reset import ForgotPasswordForm, ResetPasswordForm
from fief.dependencies.auth import get_optional_login_session
from fief.dependencies.tenant import get_current_tenant
from fief.dependencies.theme import get_current_theme
from fief.dependencies.users import UserManager, get_user_manager
from fief.forms import FormHelper
from fief.locale import gettext_lazy as _
from fief.models import LoginSession, Tenant, Theme

router = APIRouter()


@router.api_route("/forgot", methods=["GET", "POST"], name="reset:forgot")
async def forgot_password(
    request: Request,
    user_manager: UserManager = Depends(get_user_manager),
    tenant: Tenant = Depends(get_current_tenant),
    theme: Theme = Depends(get_current_theme),
):
    form_helper = FormHelper(
        ForgotPasswordForm,
        "auth/forgot_password.html",
        request=request,
        context={"tenant": tenant, "theme": theme},
    )
    form = await form_helper.get_form()

    if await form_helper.is_submitted_and_valid():
        try:
            user = await user_manager.get_by_email(form.email.data)
            await user_manager.forgot_password(user, request)
        except (UserNotExists, UserInactive):
            pass

        form_helper.context["success"] = _(
            "Check your inbox! If an account associated with this email address exists in our system, you'll receive a link to reset your password."
        )

    return await form_helper.get_response()


@router.api_route("/reset", methods=["GET", "POST"], name="reset:reset")
async def reset_password(
    request: Request,
    token: str | None = Query(None),
    user_manager: UserManager = Depends(get_user_manager),
    login_session: LoginSession | None = Depends(get_optional_login_session),
    tenant: Tenant = Depends(get_current_tenant),
    theme: Theme = Depends(get_current_theme),
):
    form_helper = FormHelper(
        ResetPasswordForm,
        "auth/reset_password.html",
        request=request,
        context={"tenant": tenant, "theme": theme},
    )
    form = await form_helper.get_form()

    if request.method == "GET":
        if token is None:
            return await form_helper.get_error_response(
                _("The reset password token is missing."), "missing_token", fatal=True
            )
        else:
            form.token.data = token

    if await form_helper.is_submitted_and_valid():
        try:
            await user_manager.reset_password(
                form.token.data, form.password.data, request
            )
        except (InvalidResetPasswordToken, UserNotExists, UserInactive):
            return await form_helper.get_error_response(
                _("The reset password token is invalid or expired."),
                "invalid_token",
                fatal=True,
            )
        except InvalidPasswordException as e:
            form.password.errors.append(e.reason)
            return await form_helper.get_error_response(e.reason, "invalid_password")
        else:
            if login_session is not None:
                redirection = tenant.url_path_for(request, "auth:login")
                return RedirectResponse(
                    url=redirection, status_code=status.HTTP_302_FOUND
                )

    return await form_helper.get_response()



from fastapi import Depends, HTTPException, Request, status

from fief.dependencies.admin_api_key import get_optional_admin_api_key
from fief.dependencies.admin_session import get_optional_admin_session_token
from fief.dependencies.current_workspace import get_current_workspace
from fief.dependencies.main_repositories import get_main_repository
from fief.models import AdminAPIKey, AdminSessionToken, Workspace
from fief.repositories import WorkspaceRepository, WorkspaceUserRepository


async def is_authenticated_admin_session(
    request: Request,
    session_token: AdminSessionToken | None = Depends(get_optional_admin_session_token),
    current_workspace: Workspace = Depends(get_current_workspace),
    workspace_user_repository: WorkspaceUserRepository = Depends(
        get_main_repository(WorkspaceUserRepository)
    ),
    workspace_repository: WorkspaceRepository = Depends(
        get_main_repository(WorkspaceRepository)
    ),
):
    if session_token is None:
        raise HTTPException(
            status_code=status.HTTP_307_TEMPORARY_REDIRECT,
            headers={"Location": request.url_for("dashboard.auth:login")},
        )

    workspace_user = await workspace_user_repository.get_by_workspace_and_user(
        current_workspace.id, session_token.user_id
    )
    if workspace_user is None:
        admin_user_workspaces = await workspace_repository.get_by_admin_user(
            session_token.user_id
        )
        if len(admin_user_workspaces) == 0:
            raise HTTPException(
                status_code=status.HTTP_307_TEMPORARY_REDIRECT,
                headers={"Location": request.url_for("dashboard.workspaces:create")},
            )
        else:
            workspace = admin_user_workspaces[0]
            redirection = f"{request.url.scheme}://{workspace.domain}/admin/"
            raise HTTPException(
                status_code=status.HTTP_307_TEMPORARY_REDIRECT,
                headers={"Location": redirection},
            )


async def is_authenticated_admin_api(
    admin_api_key: AdminAPIKey | None = Depends(get_optional_admin_api_key),
    current_workspace: Workspace = Depends(get_current_workspace),
):
    if admin_api_key is None:
        raise HTTPException(status_code=status.HTTP_401_UNAUTHORIZED)

    if admin_api_key.workspace_id != current_workspace.id:
        raise HTTPException(status_code=status.HTTP_403_FORBIDDEN)



import uuid

from sqlalchemy import select

from fief.models import WebhookLog
from fief.repositories.base import BaseRepository, UUIDRepositoryMixin


class WebhookLogRepository(BaseRepository[WebhookLog], UUIDRepositoryMixin[WebhookLog]):
    model = WebhookLog

    async def get_by_id_and_webhook(
        self, id: uuid.UUID, webhook: uuid.UUID
    ) -> WebhookLog | None:
        statement = select(WebhookLog).where(
            WebhookLog.id == id, WebhookLog.webhook_id == webhook
        )
        return await self.get_one_or_none(statement)



import base64
import hashlib
from datetime import datetime, timezone

from jwcrypto import jwk, jwt

from fief.models import Client, User


def generate_id_token(
    signing_key: jwk.JWK,
    host: str,
    client: Client,
    authenticated_at: datetime,
    user: User,
    lifetime_seconds: int,
    *,
    nonce: str | None = None,
    c_hash: str | None = None,
    access_token: str | None = None,
    encryption_key: jwk.JWK | None = None,
) -> str:
    """
    Generate an ID Token for an authenticated user.

    It's a signed JWT with claims following the OpenID specification.

    :param signing_key: The JWK to sign the JWT.
    :host: The issuer host.
    :client: The client used to authenticate the user.
    :authenticated_at: Date and time at which the user authenticated.
    :user: The authenticated user.
    :lifetime_seconds: Lifetime of the JWT.
    :nonce: Optional nonce value associated with the authorization request.
    :c_hash: Optional c_hash value to add to the claims.
    :access_token: Optional access token associated to the ID Token.
    :encryption_key: Optional JWK to further encrypt the signed token.
    In this case, it becomes a Nested JWT, as defined in rfc7519.
    """
    iat = int(datetime.now(timezone.utc).timestamp())
    exp = iat + lifetime_seconds

    claims = {
        **user.get_claims(),
        "iss": host,
        "aud": [client.client_id],
        "exp": exp,
        "iat": iat,
        "auth_time": int(authenticated_at.timestamp()),
        "azp": client.client_id,
    }

    if nonce is not None:
        claims["nonce"] = nonce
    if c_hash is not None:
        claims["c_hash"] = c_hash
    if access_token is not None:
        claims["at_hash"] = get_validation_hash(access_token)

    signed_token = jwt.JWT(
        header={"alg": "RS256", "kid": signing_key["kid"]}, claims=claims
    )
    signed_token.make_signed_token(signing_key)

    if encryption_key is not None:
        encrypted_token = jwt.JWT(
            header={
                "alg": "RSA-OAEP-256",
                "enc": "A256CBC-HS512",
                "kid": encryption_key["kid"],
            },
            claims=signed_token.serialize(),
        )
        encrypted_token.make_encrypted_token(encryption_key)
        return encrypted_token.serialize()

    return signed_token.serialize()


def get_validation_hash(value: str) -> str:
    """
    Computes a hash value to be embedded in the ID Token, like at_hash and c_hash.

    Specification: https://openid.net/specs/openid-connect-core-1_0.html#toc
    """
    hasher = hashlib.sha256()
    hasher.update(value.encode("utf-8"))
    hash = hasher.digest()

    half_hash = hash[0 : int(len(hash) / 2)]
    # Remove the Base64 padding "==" at the end
    base64_hash = base64.urlsafe_b64encode(half_hash)[:-2]

    return base64_hash.decode("utf-8")



"""Add TenantOAuthProvider

Revision ID: 30bddbb1abd8
Revises: f50cd0e11d49
Create Date: 2023-02-16 10:06:17.643794

"""
import sqlalchemy as sa
from alembic import op

import fief

# revision identifiers, used by Alembic.
revision = "30bddbb1abd8"
down_revision = "f50cd0e11d49"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "fief_tenants_oauth_providers",
        sa.Column("tenant_id", fief.models.generics.GUID(), nullable=False),
        sa.Column("oauth_provider_id", fief.models.generics.GUID(), nullable=False),
        sa.ForeignKeyConstraint(
            ["oauth_provider_id"], ["fief_oauth_providers.id"], ondelete="CASCADE"
        ),
        sa.ForeignKeyConstraint(["tenant_id"], ["fief_tenants.id"], ondelete="CASCADE"),
        sa.PrimaryKeyConstraint("tenant_id", "oauth_provider_id"),
    )

    op.execute(
        """
        INSERT INTO fief_tenants_oauth_providers
        SELECT fief_tenants.id, fief_oauth_providers.id FROM fief_tenants, fief_oauth_providers
    """
    )

    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_table("fief_tenants_oauth_providers")
    # ### end Alembic commands ###



from enum import StrEnum

from fief.services.email.base import EmailError, EmailProvider, SendEmailError
from fief.services.email.null import Null
from fief.services.email.postmark import Postmark
from fief.services.email.sendgrid import Sendgrid
from fief.services.email.smtp import SMTP


class AvailableEmailProvider(StrEnum):
    NULL = "NULL"
    POSTMARK = "POSTMARK"
    SMTP = "SMTP"
    SENDGRID = "SENDGRID"


EMAIL_PROVIDERS: dict[AvailableEmailProvider, type[EmailProvider]] = {
    AvailableEmailProvider.NULL: Null,
    AvailableEmailProvider.POSTMARK: Postmark,
    AvailableEmailProvider.SMTP: SMTP,
    AvailableEmailProvider.SENDGRID: Sendgrid,
}

__all__ = [
    "AvailableEmailProvider",
    "EMAIL_PROVIDERS",
    "EmailError",
    "EmailProvider",
    "SendEmailError",
]



from typing import Any

from fastapi import Request
from jwcrypto import jwk
from pydantic import UUID4
from sqlalchemy import Boolean, Column, ForeignKey, String, Table, Text
from sqlalchemy.orm import Mapped, mapped_column, relationship
from starlette.routing import Router

from fief.crypto.jwk import generate_signature_jwk_string, load_jwk
from fief.models.base import WorkspaceBase, get_prefixed_tablename
from fief.models.generics import GUID, CreatedUpdatedAt, UUIDModel
from fief.models.oauth_provider import OAuthProvider
from fief.models.theme import Theme

TenantOAuthProvider = Table(
    get_prefixed_tablename("tenants_oauth_providers"),
    WorkspaceBase.metadata,
    Column(
        "tenant_id",
        ForeignKey(f"{get_prefixed_tablename('tenants')}.id", ondelete="CASCADE"),
        primary_key=True,
    ),
    Column(
        "oauth_provider_id",
        ForeignKey(
            f"{get_prefixed_tablename('oauth_providers')}.id", ondelete="CASCADE"
        ),
        primary_key=True,
    ),
)


class Tenant(UUIDModel, CreatedUpdatedAt, WorkspaceBase):
    __tablename__ = "tenants"

    name: Mapped[str] = mapped_column(String(length=255), nullable=False)
    slug: Mapped[str] = mapped_column(String(length=255), nullable=False, unique=True)
    default: Mapped[bool] = mapped_column(Boolean, default=False, nullable=False)
    sign_jwk: Mapped[str] = mapped_column(
        Text, nullable=False, default=generate_signature_jwk_string
    )
    registration_allowed: Mapped[bool] = mapped_column(
        Boolean, default=True, nullable=False
    )

    theme_id: Mapped[UUID4 | None] = mapped_column(
        GUID, ForeignKey(Theme.id, ondelete="SET NULL"), nullable=True
    )
    theme: Mapped[Theme | None] = relationship("Theme")

    logo_url: Mapped[str | None] = mapped_column(
        String(length=512), default=None, nullable=True
    )

    oauth_providers: Mapped[list[OAuthProvider]] = relationship(
        "OAuthProvider", secondary=TenantOAuthProvider, lazy="selectin"
    )

    def __repr__(self) -> str:
        return f"Tenant(id={self.id}, name={self.name}, slug={self.slug}, default={self.default})"

    def get_sign_jwk(self) -> jwk.JWK:
        return load_jwk(self.sign_jwk)

    def get_host(self, domain: str) -> str:
        host = f"https://{domain}"
        if not self.default:
            host += f"/{self.slug}"
        return host

    def url_for(self, request: Request, name: str, **path_params: Any) -> str:
        if not self.default:
            path_params["tenant_slug"] = self.slug
        return request.url_for(name, **path_params)

    def url_path_for(self, request: Request, name: str, **path_params: Any) -> str:
        if not self.default:
            path_params["tenant_slug"] = self.slug
        router: Router = request.scope["router"]
        return str(router.url_path_for(name, **path_params))

    def get_oauth_provider(self, id: UUID4) -> OAuthProvider | None:
        for oauth_provider in self.oauth_providers:
            if oauth_provider.id == id:
                return oauth_provider
        return None



from sqlalchemy import select

from fief.models import EmailTemplate
from fief.repositories.base import BaseRepository, UUIDRepositoryMixin
from fief.services.email_template.types import EmailTemplateType


class EmailTemplateRepository(
    BaseRepository[EmailTemplate], UUIDRepositoryMixin[EmailTemplate]
):
    model = EmailTemplate

    async def get_by_type(self, type: EmailTemplateType) -> EmailTemplate | None:
        statement = select(EmailTemplate).where(EmailTemplate.type == type.value)
        return await self.get_one_or_none(statement)



from pydantic import UUID4
from sqlalchemy import JSON, ForeignKey
from sqlalchemy.orm import Mapped, mapped_column, relationship
from sqlalchemy.sql.schema import UniqueConstraint

from fief.models.base import WorkspaceBase
from fief.models.client import Client
from fief.models.generics import GUID, CreatedUpdatedAt, UUIDModel
from fief.models.user import User


class Grant(UUIDModel, CreatedUpdatedAt, WorkspaceBase):
    __tablename__ = "grants"
    __table_args__ = (UniqueConstraint("user_id", "client_id"),)

    scope: Mapped[list[str]] = mapped_column(JSON, nullable=False, default=list)

    user_id: Mapped[UUID4] = mapped_column(
        GUID, ForeignKey(User.id, ondelete="CASCADE"), nullable=False
    )
    user: Mapped[User] = relationship("User")

    client_id: Mapped[UUID4] = mapped_column(
        GUID, ForeignKey(Client.id, ondelete="CASCADE"), nullable=False
    )
    client: Mapped[Client] = relationship("Client", lazy="joined")



import functools
import http.cookies

from starlette.datastructures import MutableHeaders
from starlette.types import ASGIApp, Message, Receive, Scope, Send

from fief.settings import settings

CSRF_ATTRIBUTE_NAME = "csrftoken"
SAFE_METHODS = {"GET", "HEAD", "OPTIONS", "TRACE"}


class CSRFCookieSetterMiddleware:
    def __init__(self, app: ASGIApp):
        self.app = app

    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
        if scope["type"] != "http":  # pragma: no cover
            await self.app(scope, receive, send)
            return

        send = functools.partial(self.send, send=send, scope=scope)
        await self.app(scope, receive, send)

    async def send(self, message: Message, send: Send, scope: Scope) -> None:
        if message["type"] != "http.response.start":
            await send(message)
            return

        if csrftoken := scope.get(CSRF_ATTRIBUTE_NAME):
            message.setdefault("headers", [])
            headers = MutableHeaders(scope=message)

            cookie: http.cookies.BaseCookie = http.cookies.SimpleCookie()
            cookie_name = settings.csrf_cookie_name
            cookie[cookie_name] = csrftoken
            cookie[cookie_name]["secure"] = settings.csrf_cookie_secure
            cookie[cookie_name]["httponly"] = True
            headers.append("set-cookie", cookie.output(header="").strip())

        await send(message)



from sqlalchemy.orm import DeclarativeBase

from fief.settings import settings


class MainBase(DeclarativeBase):
    pass


def get_prefixed_tablename(name: str) -> str:
    return f"{settings.workspace_table_prefix}{name}"


class WorkspaceBase(DeclarativeBase):
    def __init_subclass__(cls) -> None:
        cls.__tablename__ = get_prefixed_tablename(cls.__tablename__)
        super().__init_subclass__()



import random
import string

from slugify import slugify
from sqlalchemy import select

from fief.models import Tenant
from fief.repositories.base import BaseRepository, UUIDRepositoryMixin


class TenantRepository(BaseRepository[Tenant], UUIDRepositoryMixin[Tenant]):
    model = Tenant

    async def get_default(self) -> Tenant | None:
        statement = select(Tenant).where(Tenant.default == True)
        return await self.get_one_or_none(statement)

    async def get_by_slug(self, slug: str) -> Tenant | None:
        statement = select(Tenant).where(Tenant.slug == slug)
        return await self.get_one_or_none(statement)

    async def get_available_slug(self, name: str) -> str:
        slug = slugify(name)
        tenant = await self.get_by_slug(slug)

        if tenant is None:
            return slug

        random_string = "".join(
            random.choices(string.ascii_lowercase + string.digits, k=6)
        )
        return f"{slug}-{random_string}"



from wtforms import StringField, TextAreaField

from fief.forms import CSRFBaseForm


class EmailTemplateUpdateForm(CSRFBaseForm):
    subject = StringField("Subject")
    content = TextAreaField("Content")



from wtforms import (
    IntegerField,
    PasswordField,
    RadioField,
    SelectField,
    StringField,
    validators,
    widgets,
)

from fief.db.types import SSL_MODES, DatabaseType
from fief.forms import CSRFBaseForm


class WorkspaceCreateStep1Form(CSRFBaseForm):
    name = StringField(
        "Name of your workspace", validators=[validators.InputRequired()]
    )


class WorkspaceCreateStep2Form(CSRFBaseForm):
    database = RadioField(
        choices=[("cloud", "cloud"), ("custom", "custom")],
        validators=[validators.InputRequired()],
    )


class WorkspaceCreateStep3Form(CSRFBaseForm):
    database_type = SelectField(
        label="Database type",
        choices=DatabaseType.choices(),
        coerce=DatabaseType.coerce,
        default=DatabaseType.POSTGRESQL.value,
        validators=[validators.InputRequired()],
    )
    database_host = StringField(label="Host", validators=[validators.InputRequired()])
    database_port = IntegerField(label="Port", validators=[validators.InputRequired()])
    database_username = StringField(
        label="Username", validators=[validators.InputRequired()]
    )
    database_password = PasswordField(
        label="Password",
        validators=[validators.InputRequired()],
        widget=widgets.PasswordInput(hide_value=False),
    )
    database_name = StringField(
        label="Database name", validators=[validators.InputRequired()]
    )
    database_ssl_mode = SelectField(
        label="SSL mode", validators=[validators.InputRequired()]
    )

    def prefill_from_database_type(self):
        database_type = DatabaseType[self.database_type.data]
        ssl_modes_enum = SSL_MODES[database_type]
        self.database_ssl_mode.choices = ssl_modes_enum.choices()
        if not self.database_ssl_mode.data:
            self.database_ssl_mode.process_data(self.database_ssl_mode.choices[0][0])


class WorkspaceCreateStep4Form(CSRFBaseForm):
    pass



from sqlalchemy import String, Text
from sqlalchemy.orm import Mapped, mapped_column

from fief.models.base import WorkspaceBase
from fief.models.generics import CreatedUpdatedAt, UUIDModel
from fief.services.email_template.types import EmailTemplateType


class EmailTemplate(UUIDModel, CreatedUpdatedAt, WorkspaceBase):
    __tablename__ = "email_templates"

    type: Mapped[EmailTemplateType] = mapped_column(
        String(length=255), nullable=False, unique=True
    )
    subject: Mapped[str] = mapped_column(Text, nullable=False)
    content: Mapped[str] = mapped_column(Text, nullable=False)

    def get_type_display_name(self) -> str:
        return EmailTemplateType[self.type].get_display_name()



import secrets
from datetime import datetime, timedelta, timezone
from enum import StrEnum

from jwcrypto import jwk
from pydantic import UUID4
from sqlalchemy import Boolean, Enum, ForeignKey, Integer, String, Text
from sqlalchemy.orm import Mapped, mapped_column, relationship
from sqlalchemy.sql.sqltypes import JSON

from fief.crypto.jwk import load_jwk
from fief.models.base import WorkspaceBase
from fief.models.generics import GUID, CreatedUpdatedAt, UUIDModel
from fief.models.tenant import Tenant
from fief.settings import settings


def get_default_redirect_uris() -> list[str]:
    return ["http://localhost:8000/docs/oauth2-redirect"]


class ClientType(StrEnum):
    PUBLIC = "public"
    CONFIDENTIAL = "confidential"

    def get_display_name(self) -> str:
        display_names = {
            ClientType.PUBLIC: "Public",
            ClientType.CONFIDENTIAL: "Confidential",
        }
        return display_names[self]

    @classmethod
    def choices(cls) -> list[tuple[str, str]]:
        return [(member.value, member.get_display_name()) for member in cls]

    @classmethod
    def coerce(cls, item):
        return cls(str(item)) if not isinstance(item, cls) else item


class Client(UUIDModel, CreatedUpdatedAt, WorkspaceBase):
    __tablename__ = "clients"

    name: Mapped[str] = mapped_column(String(length=255), nullable=False)
    first_party: Mapped[bool] = mapped_column(Boolean, nullable=False, default=False)

    client_type: Mapped[ClientType] = mapped_column(
        Enum(ClientType, values_callable=lambda x: [e.value for e in x]),
        nullable=False,
        default=ClientType.CONFIDENTIAL,
        server_default=ClientType.CONFIDENTIAL,
    )
    client_id: Mapped[str] = mapped_column(
        String(length=255), default=secrets.token_urlsafe, nullable=False, index=True
    )
    client_secret: Mapped[str] = mapped_column(
        String(length=255), default=secrets.token_urlsafe, nullable=False, index=True
    )
    redirect_uris: Mapped[list[str]] = mapped_column(
        JSON, nullable=False, default=get_default_redirect_uris
    )
    encrypt_jwk: Mapped[str] = mapped_column(Text, nullable=True)

    authorization_code_lifetime_seconds: Mapped[int] = mapped_column(
        Integer,
        nullable=False,
        default=settings.default_authorization_code_lifetime_seconds,
    )
    access_id_token_lifetime_seconds: Mapped[int] = mapped_column(
        Integer,
        nullable=False,
        default=settings.default_access_id_token_lifetime_seconds,
    )
    refresh_token_lifetime_seconds: Mapped[int] = mapped_column(
        Integer, nullable=False, default=settings.default_refresh_token_lifetime_seconds
    )

    tenant_id: Mapped[UUID4] = mapped_column(
        GUID, ForeignKey(Tenant.id, ondelete="CASCADE"), nullable=False
    )
    tenant: Mapped[Tenant] = relationship("Tenant", lazy="joined")

    def __repr__(self) -> str:
        return f"Client(id={self.id}, name={self.name}, client_id={self.client_id})"

    def get_encrypt_jwk(self) -> jwk.JWK | None:
        if self.encrypt_jwk is None:
            return None
        return load_jwk(self.encrypt_jwk)

    def get_authorization_code_expires_at(self) -> datetime:
        return self._get_expires_at("authorization_code_lifetime_seconds")

    def get_access_id_token_expires_at(self) -> datetime:
        return self._get_expires_at("access_id_token_lifetime_seconds")

    def get_refresh_token_expires_at(self) -> datetime:
        return self._get_expires_at("refresh_token_lifetime_seconds")

    def _get_expires_at(self, attr: str) -> datetime:
        return datetime.now(timezone.utc) + timedelta(seconds=getattr(self, attr))



from wtforms import BooleanField, StringField, URLField, validators

from fief.forms import (
    ComboboxSelectField,
    ComboboxSelectMultipleField,
    CSRFBaseForm,
    empty_string_to_none,
)


class BaseTenantForm(CSRFBaseForm):
    name = StringField("Name", validators=[validators.InputRequired()])
    registration_allowed = BooleanField("Registration allowed", default=True)
    logo_url = URLField(
        "Logo URL",
        validators=[validators.Optional(), validators.URL(require_tld=False)],
        filters=[empty_string_to_none],
        description="It will be shown on the top left of authentication pages.",
    )
    theme = ComboboxSelectField(
        "UI Theme",
        query_endpoint_path="/admin/customization/themes/",
        validators=[validators.Optional(), validators.UUID()],
        filters=[empty_string_to_none],
        description="If left empty, the default theme will be used.",
    )
    oauth_providers = ComboboxSelectMultipleField(
        "OAuth Providers",
        query_endpoint_path="/admin/oauth-providers/",
        label_attr="display_name",
        choices=[],
        validate_choice=False,
        description="OAuth Providers users should be allowed to use to login.",
    )


class TenantCreateForm(BaseTenantForm):
    pass


class TenantUpdateForm(BaseTenantForm):
    pass



from typing import Any

from fastapi import APIRouter, Depends, Header, Request, status
from fastapi.responses import RedirectResponse
from pydantic import ValidationError

from fief.apps.dashboard.forms.workspace import (
    WorkspaceCreateStep1Form,
    WorkspaceCreateStep2Form,
    WorkspaceCreateStep3Form,
    WorkspaceCreateStep4Form,
)
from fief.db.types import UNSAFE_SSL_MODES, create_database_connection_parameters
from fief.dependencies.admin_session import get_admin_session_token
from fief.dependencies.workspace_creation import get_workspace_creation
from fief.dependencies.workspace_db import get_workspace_db
from fief.forms import FormHelper
from fief.models import AdminSessionToken
from fief.schemas.workspace import WorkspaceCreate
from fief.services.workspace_creation import WorkspaceCreation
from fief.services.workspace_db import (
    WorkspaceDatabase,
    WorkspaceDatabaseConnectionError,
)

CREATE_WORKSPACE_DATA_SESSION_KEY = "create_workspace"
CREATE_WORKSPACE_NB_STEPS = 4

router = APIRouter(dependencies=[Depends(get_admin_session_token)])


async def get_base_context(request: Request, hx_request: bool = Header(False)):
    layout = "admin/workspaces/create/layout.html"
    if hx_request:
        layout = "admin/workspaces/create/layout_boost.html"
    return {"request": request, "layout": layout, "steps": CREATE_WORKSPACE_NB_STEPS}


async def get_create_workspace_session_data(request: Request) -> dict[str, Any]:
    return request.session.get(CREATE_WORKSPACE_DATA_SESSION_KEY, {})


@router.get("/create", name="dashboard.workspaces:create")
async def create_workspace(request: Request):
    return RedirectResponse(request.url_for("dashboard.workspaces:create.step1"))


@router.api_route(
    "/create/step1", methods=["GET", "POST"], name="dashboard.workspaces:create.step1"
)
async def create_workspace_step1(
    request: Request,
    session_data: dict[str, Any] = Depends(get_create_workspace_session_data),
    context=Depends(get_base_context),
):
    form_helper = FormHelper(
        WorkspaceCreateStep1Form,
        "admin/workspaces/create/step1.html",
        data=session_data,
        request=request,
        context={**context, "active": 0},
    )

    if await form_helper.is_submitted_and_valid():
        form = await form_helper.get_form()
        data = form.data
        data.pop("csrf_token")

        request.session[CREATE_WORKSPACE_DATA_SESSION_KEY] = {
            **session_data,
            **data,
        }

        return RedirectResponse(
            request.url_for("dashboard.workspaces:create.step2"),
            status_code=status.HTTP_303_SEE_OTHER,
        )

    return await form_helper.get_response()


@router.api_route(
    "/create/step2", methods=["GET", "POST"], name="dashboard.workspaces:create.step2"
)
async def create_workspace_step2(
    request: Request,
    context=Depends(get_base_context),
):
    form_helper = FormHelper(
        WorkspaceCreateStep2Form,
        "admin/workspaces/create/step2.html",
        request=request,
        context={**context, "active": 1},
    )

    if await form_helper.is_submitted_and_valid():
        form = await form_helper.get_form()
        data = form.data

        redirect_url = request.url_for("dashboard.workspaces:create.step3")
        if data["database"] == "cloud":
            redirect_url = request.url_for("dashboard.workspaces:create.step4")

        return RedirectResponse(redirect_url, status_code=status.HTTP_303_SEE_OTHER)

    return await form_helper.get_response()


@router.api_route(
    "/create/step3", methods=["GET", "POST"], name="dashboard.workspaces:create.step3"
)
async def create_workspace_step3(
    request: Request,
    hx_trigger: str | None = Header(None),
    session_data: dict[str, Any] = Depends(get_create_workspace_session_data),
    context=Depends(get_base_context),
):
    form_helper = FormHelper(
        WorkspaceCreateStep3Form,
        "admin/workspaces/create/step3.html",
        request=request,
        data=session_data,
        context={**context, "active": 2, "UNSAFE_SSL_MODES": UNSAFE_SSL_MODES},
    )
    form = await form_helper.get_form()
    form.prefill_from_database_type()

    if hx_trigger is None and await form_helper.is_submitted_and_valid():
        data = form.data
        data.pop("csrf_token")

        request.session[CREATE_WORKSPACE_DATA_SESSION_KEY] = {
            **session_data,
            **data,
        }

        return RedirectResponse(
            request.url_for("dashboard.workspaces:create.step4"),
            status_code=status.HTTP_303_SEE_OTHER,
        )

    return await form_helper.get_response()


@router.post(
    "/create/step3/check-connection",
    name="dashboard.workspaces:create.step3.check_connection",
)
async def create_workspace_step3_check_connection(
    request: Request,
    context=Depends(get_base_context),
    workspace_db: WorkspaceDatabase = Depends(get_workspace_db),
):
    form_helper = FormHelper(
        WorkspaceCreateStep3Form,
        "admin/workspaces/create/step3.html",
        request=request,
        context={**context, "active": 2, "UNSAFE_SSL_MODES": UNSAFE_SSL_MODES},
    )
    form = await form_helper.get_form()
    form.prefill_from_database_type()

    if await form_helper.is_submitted_and_valid():
        data = form.data

        database_connection_parameters = create_database_connection_parameters(
            data["database_type"],
            asyncio=False,
            username=data["database_username"],
            password=data["database_password"],
            host=data["database_host"],
            port=data["database_port"],
            database=data["database_name"],
            ssl_mode=data["database_ssl_mode"],
        )
        valid, message = workspace_db.check_connection(database_connection_parameters)
        if not valid:
            return await form_helper.get_error_response(
                message or "Unknown error", "database_connection_error"
            )
        else:
            form_helper.context.update({"database_connection_success": True})

    return await form_helper.get_response()


@router.api_route(
    "/create/step4", methods=["GET", "POST"], name="dashboard.workspaces:create.step4"
)
async def create_workspace_step4(
    request: Request,
    session_data: dict[str, Any] = Depends(get_create_workspace_session_data),
    workspace_creation: WorkspaceCreation = Depends(get_workspace_creation),
    admin_session_token: AdminSessionToken = Depends(get_admin_session_token),
    context=Depends(get_base_context),
):
    form_helper = FormHelper(
        WorkspaceCreateStep4Form,
        "admin/workspaces/create/step4.html",
        request=request,
        context={**context, "active": 3},
    )

    if await form_helper.is_submitted_and_valid():
        try:
            workspace_create = WorkspaceCreate(**session_data)
            workspace = await workspace_creation.create(
                workspace_create, user_id=admin_session_token.user_id
            )
        except ValidationError:
            request.session[CREATE_WORKSPACE_DATA_SESSION_KEY] = {}
            return RedirectResponse(
                request.url_for("dashboard.workspaces:create"),
                status_code=status.HTTP_303_SEE_OTHER,
            )
        except WorkspaceDatabaseConnectionError as e:
            return await form_helper.get_error_response(
                e.message, "database_connection_error"
            )

        request.session[CREATE_WORKSPACE_DATA_SESSION_KEY] = {}
        return RedirectResponse(
            f"{request.url.scheme}://{workspace.domain}/admin/",
            status_code=status.HTTP_303_SEE_OTHER,
        )

    return await form_helper.get_response()



import smtplib
import ssl
from email.message import EmailMessage

from fief.services.email.base import EmailProvider, SendEmailError, format_address


class SMTP(EmailProvider):
    def __init__(
        self,
        host: str,
        username: str | None = None,
        password: str | None = None,
        port: int = 587,
        ssl: bool | None = True,
    ) -> None:
        self.username = username
        self.password = password
        self.host = host
        self.port = port
        self.ssl = ssl

    def send_email(
        self,
        *,
        sender: tuple[str, str | None],
        recipient: tuple[str, str | None],
        subject: str,
        html: str | None = None,
        text: str | None = None,
    ):
        from_email, from_name = sender
        to_email, to_name = recipient

        try:
            message = EmailMessage()
            message["Subject"] = subject
            message["From"] = format_address(from_email, from_name)
            message["To"] = format_address(to_email, to_name)
            if html is not None:
                message.add_alternative(html, subtype="html")
            if text is not None:
                message.add_alternative(text, subtype="plain")

            with smtplib.SMTP(self.host, self.port) as server:
                if self.ssl:
                    context = ssl.create_default_context()
                    server.starttls(context=context)
                if self.username and self.password:
                    server.login(self.username, self.password)
                server.send_message(message)
        except smtplib.SMTPException as e:
            raise SendEmailError(str(e)) from e



import hmac
import time
from hashlib import sha256

import httpx

from fief import __version__
from fief.models import Webhook, WebhookLog
from fief.repositories import WebhookLogRepository
from fief.services.webhooks.models import WebhookEvent


class WebhookDeliveryError(Exception):
    def __init__(self, message: str) -> None:
        super().__init__(message)


class WebhookDelivery:
    def __init__(self, webhook_log_repository: WebhookLogRepository) -> None:
        self.webhook_log_repository = webhook_log_repository

    async def deliver(self, webhook: Webhook, event: WebhookEvent, attempt: int = 1):
        async with httpx.AsyncClient() as client:
            payload = event.json()
            signature, ts = self._get_signature(payload, webhook.secret)

            webhook_log = WebhookLog(
                webhook_id=webhook.id,
                event=event.type,
                attempt=attempt,
                payload=payload,
                success=False,
            )

            try:
                response = await client.post(
                    webhook.url,
                    content=payload,
                    headers={
                        "User-Agent": f"fief-server-webhooks/{__version__}",
                        "Content-Type": "application/json",
                        "X-Fief-Webhook-Signature": signature,
                        "X-Fief-Webhook-Timestamp": str(ts),
                    },
                    follow_redirects=False,
                )
                webhook_log.response = response.text
                response.raise_for_status()
                webhook_log.success = True
            except httpx.HTTPError as e:
                webhook_log.error_type = type(e).__name__
                webhook_log.error_message = str(e)
                raise WebhookDeliveryError(str(e)) from e
            finally:
                await self.webhook_log_repository.create(webhook_log)

    def _get_signature(self, payload: str, secret: str) -> tuple[str, int]:
        ts = int(time.time())
        message = f"{ts}.{payload}"

        hash = hmac.new(
            secret.encode("utf-8"),
            msg=message.encode("utf-8"),
            digestmod=sha256,
        )
        signature = hash.hexdigest()
        return signature, ts



from pydantic import AnyUrl

from fief.schemas.generics import BaseModel


class OAuth2AuthorizationServerMetadata(BaseModel):
    """
    OAuth 2.0 Authorization Server Metadata

    Conforms to RFC8414.
    https://datatracker.ietf.org/doc/html/rfc8414
    """

    issuer: AnyUrl
    authorization_endpoint: AnyUrl
    token_endpoint: AnyUrl
    jwks_uri: AnyUrl
    registration_endpoint: AnyUrl
    scopes_supported: list[str]
    response_types_supported: list[str]
    response_modes_supported: list[str] | None = None
    grant_types_supported: list[str] | None = None
    token_endpoint_auth_methods_supported: list[str] | None = None
    token_endpoint_auth_signing_alg_values_supported: list[str] | None = None
    service_documentation: AnyUrl | None = None
    ui_locales_supported: list[str] | None = None
    op_policy_uri: AnyUrl | None = None
    op_tos_uri: AnyUrl | None = None
    revocation_endpoint: AnyUrl | None = None
    revocation_endpoint_auth_methods_supported: list[str] | None = None
    revocation_endpoint_auth_signing_alg_values_supported: list[str] | None = None
    introspection_endpoint: AnyUrl | None = None
    introspection_endpoint_auth_methods_supported: list[str] | None = None
    introspection_endpoint_auth_signing_alg_values_supported: list[str] | None = None
    code_challenge_methods_supported: list[str] | None = None


class OpenIDProviderMetadata(OAuth2AuthorizationServerMetadata):
    """
    OpenID Provider Metadata

    Conforms to OpenID Connect Discovery 1.0 specification.
    https://openid.net/specs/openid-connect-discovery-1_0.html#ProviderMetadata
    """

    userinfo_endpoint: AnyUrl
    acr_values_supported: list[str] | None = None
    subject_types_supported: list[str]
    id_token_signing_alg_values_supported: list[str]
    id_token_encryption_alg_values_supported: list[str] | None = None
    id_token_encryption_enc_values_supported: list[str] | None = None
    userinfo_signing_alg_values_supported: list[str] | None = None
    userinfo_encryption_alg_values_supported: list[str] | None = None
    userinfo_encryption_enc_values_supported: list[str] | None = None
    request_object_signing_alg_values_supported: list[str] | None = None
    request_object_encryption_alg_values_supported: list[str] | None = None
    request_object_encryption_enc_values_supported: list[str] | None = None
    token_endpoint_auth_methods_supported: list[str] | None = None
    token_endpoint_auth_signing_alg_values_supported: list[str] | None = None
    display_values_supported: list[str] | None = None
    claim_types_supported: list[str] | None = None
    claims_supported: list[str] | None = None
    claims_locales_supported: list[str] | None = None
    claims_parameter_supported: bool | None = None
    request_parameter_supported: bool | None = None
    request_uri_parameter_supported: bool | None = None
    require_request_uri_registration: bool | None = None



"""Add AuditLog model to workspaces

Revision ID: f3784e0c8524
Revises: 46fa2e6c46f7
Create Date: 2022-06-20 09:12:14.136072

"""
import sqlalchemy as sa
from alembic import op

import fief

# revision identifiers, used by Alembic.
revision = "f3784e0c8524"
down_revision = "46fa2e6c46f7"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "fief_audit_logs",
        sa.Column("id", fief.models.generics.GUID(), nullable=False),
        sa.Column(
            "timestamp",
            fief.models.generics.TIMESTAMPAware(timezone=True),
            nullable=False,
        ),
        sa.Column("level", sa.String(length=255), nullable=False),
        sa.Column("message", sa.Text(), nullable=False),
        sa.Column("extra", sa.JSON(), nullable=True),
        sa.Column("author_user_id", fief.models.generics.GUID(), nullable=True),
        sa.Column("subject_user_id", fief.models.generics.GUID(), nullable=True),
        sa.ForeignKeyConstraint(
            ["author_user_id"], ["fief_users.id"], ondelete="SET NULL"
        ),
        sa.ForeignKeyConstraint(
            ["subject_user_id"], ["fief_users.id"], ondelete="SET NULL"
        ),
        sa.PrimaryKeyConstraint("id"),
    )
    op.create_index(
        op.f("ix_fief_audit_logs_level"), "fief_audit_logs", ["level"], unique=False
    )
    op.create_index(
        op.f("ix_fief_audit_logs_timestamp"),
        "fief_audit_logs",
        ["timestamp"],
        unique=False,
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_index(op.f("ix_fief_audit_logs_timestamp"), table_name="fief_audit_logs")
    op.drop_index(op.f("ix_fief_audit_logs_level"), table_name="fief_audit_logs")
    op.drop_table("fief_audit_logs")
    # ### end Alembic commands ###



from fief.services.email.base import EmailProvider


class Null(EmailProvider):
    def send_email(
        self,
        *,
        sender: tuple[str, str | None],
        recipient: tuple[str, str | None],
        subject: str,
        html: str | None = None,
        text: str | None = None,
    ):
        return



from typing import TYPE_CHECKING, Literal, overload

import jinja2

from fief.services.email_template.types import EmailTemplateType

if TYPE_CHECKING:  # pragma: no cover
    from fief.models import EmailTemplate
    from fief.repositories import EmailTemplateRepository
    from fief.services.email_template.contexts import (
        EmailContext,
        ForgotPasswordContext,
        WelcomeContext,
    )


def _templates_list_to_map(
    templates: list["EmailTemplate"],
) -> dict[str, "EmailTemplate"]:
    return {template.type: template for template in templates}


class EmailTemplateLoader:
    def __init__(
        self,
        templates: list["EmailTemplate"],
        *,
        templates_overrides: dict[EmailTemplateType, "EmailTemplate"] | None = None,
    ):
        self.templates_map = _templates_list_to_map(templates)
        if templates_overrides:
            for type, template in templates_overrides.items():
                self.templates_map[type] = template

    def __call__(self, name: str) -> str:
        return self.templates_map[name].content


class EmailTemplateRenderer:
    def __init__(
        self,
        repository: "EmailTemplateRepository",
        *,
        templates_overrides: dict[EmailTemplateType, "EmailTemplate"] | None = None,
    ):
        self.repository = repository
        self._jinja_environment: jinja2.Environment | None = None
        self.templates_overrides = templates_overrides

    @overload
    async def render(
        self, type: Literal[EmailTemplateType.WELCOME], context: "WelcomeContext"
    ) -> str:
        ...  # pragma: no cover

    @overload
    async def render(
        self,
        type: Literal[EmailTemplateType.FORGOT_PASSWORD],
        context: "ForgotPasswordContext",
    ) -> str:
        ...  # pragma: no cover

    async def render(self, type, context: "EmailContext") -> str:
        jinja_environment = await self._get_jinja_environment()
        template_object = jinja_environment.get_template(type.value)
        return template_object.render(context.dict())

    async def _get_jinja_environment(self) -> jinja2.Environment:
        if self._jinja_environment is None:
            templates = await self.repository.all()
            loader = jinja2.FunctionLoader(
                EmailTemplateLoader(
                    templates, templates_overrides=self.templates_overrides
                )
            )
            self._jinja_environment = jinja2.Environment(loader=loader, autoescape=True)
        return self._jinja_environment


class EmailSubjectLoader:
    def __init__(
        self,
        templates: list["EmailTemplate"],
        *,
        templates_overrides: dict[EmailTemplateType, "EmailTemplate"] | None = None,
    ):
        self.templates_map = _templates_list_to_map(templates)
        if templates_overrides:
            for type, template in templates_overrides.items():
                self.templates_map[type] = template

    def __call__(self, name: str) -> str:
        return self.templates_map[name].subject


class EmailSubjectRenderer:
    def __init__(
        self,
        repository: "EmailTemplateRepository",
        *,
        templates_overrides: dict[EmailTemplateType, "EmailTemplate"] | None = None,
    ):
        self.repository = repository
        self._jinja_environment: jinja2.Environment | None = None
        self.templates_overrides = templates_overrides

    @overload
    async def render(
        self, type: Literal[EmailTemplateType.WELCOME], context: "WelcomeContext"
    ) -> str:
        ...  # pragma: no cover

    @overload
    async def render(
        self,
        type: Literal[EmailTemplateType.FORGOT_PASSWORD],
        context: "ForgotPasswordContext",
    ) -> str:
        ...  # pragma: no cover

    async def render(self, type, context: "EmailContext") -> str:
        jinja_environment = await self._get_jinja_environment()
        subject_template_object = jinja_environment.get_template(type.value)
        return subject_template_object.render(context.dict())

    async def _get_jinja_environment(self) -> jinja2.Environment:
        if self._jinja_environment is None:
            templates = await self.repository.all()
            loader = jinja2.FunctionLoader(
                EmailSubjectLoader(
                    templates, templates_overrides=self.templates_overrides
                )
            )
            self._jinja_environment = jinja2.Environment(loader=loader, autoescape=True)
        return self._jinja_environment



from collections.abc import Sequence
from datetime import datetime, timezone
from typing import Any, Generic, Protocol, TypeVar, cast

from pydantic import UUID4
from sqlalchemy import delete, func, select
from sqlalchemy.engine import Result
from sqlalchemy.ext.asyncio import AsyncSession
from sqlalchemy.orm import InstrumentedAttribute, RelationshipProperty, contains_eager
from sqlalchemy.sql import Executable, Select

from fief.models.generics import M_EXPIRES_AT, M_UUID, M


class BaseRepositoryProtocol(Protocol[M]):
    model: type[M]
    session: AsyncSession

    def __init__(self, session: AsyncSession) -> None:
        ...

    async def paginate(
        self,
        statement: Select,
        limit=10,
        skip=0,
    ) -> tuple[list[M], int]:
        ...  # pragma: no cover

    def orderize(
        self, statement: Select, ordering: list[tuple[list[str], bool]]
    ) -> Select:
        ...  # pragma: no cover

    async def get_one_or_none(self, statement: Select) -> M | None:
        ...  # pragma: no cover

    async def list(self, statement: Select) -> list[M]:
        ...  # pragma: no cover

    async def create(self, object: M) -> M:
        ...  # pragma: no cover

    async def update(self, object: M) -> None:
        ...  # pragma: no cover

    async def delete(self, object: M) -> None:
        ...  # pragma: no cover

    async def _execute_query(self, statement: Select) -> Result:
        ...  # pragma: no cover

    async def _execute_statement(self, statement: Executable) -> Result:
        ...  # pragma: no cover


class UUIDRepositoryProtocol(BaseRepositoryProtocol, Protocol[M_UUID]):
    model: type[M_UUID]

    async def get_by_id(self, id: UUID4) -> M_UUID | None:
        ...  # pragma: no cover


class ExpiresAtRepositoryProtocol(BaseRepositoryProtocol, Protocol[M_EXPIRES_AT]):
    model: type[M_EXPIRES_AT]

    async def delete_expired(self) -> None:
        ...  # pragma: no cover


class BaseRepository(BaseRepositoryProtocol, Generic[M]):
    def __init__(self, session: AsyncSession) -> None:
        self.session = session

    async def paginate(
        self,
        statement: Select,
        limit=10,
        skip=0,
    ) -> tuple[list[M], int]:
        paginated_statement = statement.offset(skip).limit(limit)
        # FIXME: running it concurrently causes issues with SQLite and SQLAlchemy2
        count = await self._count(statement)
        results = await self.list(paginated_statement)
        return results, count

    def orderize(
        self, statement: Select, ordering: list[tuple[list[str], bool]]
    ) -> Select:
        for accessors, is_desc in ordering:
            field: InstrumentedAttribute
            # Local field
            if len(accessors) == 1:
                try:
                    field = getattr(self.model, accessors[0])
                    if not isinstance(
                        field.prop, RelationshipProperty
                    ):  # Prevent ordering by raw relation ship field -> "tenant" instead of "tenant.id"
                        statement = statement.order_by(
                            field.desc() if is_desc else field.asc()
                        )
                except AttributeError:
                    pass
            # Relationship field
            else:
                valid_field = True
                model = self.model
                for accessor in accessors:
                    try:
                        field = getattr(model, accessor)
                        if isinstance(field.prop, RelationshipProperty):
                            statement = statement.join(field)
                            if field.prop.lazy == "joined":
                                # If the relationship is eagerly loaded,
                                # this ensures we reuse the existing JOIN instead of adding another
                                # Ref: https://docs.sqlalchemy.org/en/14/orm/loading_relationships.html#routing-explicit-joins-statements-into-eagerly-loaded-collections
                                statement = statement.options(contains_eager(field))
                            model = field.prop.entity.class_
                    except AttributeError:
                        valid_field = False
                        break
                if valid_field:
                    statement = statement.order_by(
                        field.desc() if is_desc else field.asc()
                    )
        return statement

    async def all(self) -> list[M]:
        return await self.list(select(self.model))

    async def get_one_or_none(self, statement: Select) -> M | None:
        result = await self._execute_query(statement)
        return result.scalar()

    async def create(self, object: M) -> M:
        self.session.add(object)
        await self.session.commit()
        return object

    async def update(self, object: M) -> None:
        self.session.add(object)
        await self.session.commit()

    async def delete(self, object: M) -> None:
        await self.session.delete(object)
        await self.session.commit()

    async def create_many(self, objects: list[M]) -> list[M]:
        self.session.add_all(objects)
        await self.session.commit()
        return objects

    async def list(self, statement: Select) -> list[M]:
        result = await self._execute_query(statement)
        return cast(list[M], result.scalars().unique().all())

    async def _count(self, statement: Select) -> int:
        count_statement = statement.with_only_columns(
            func.count(), maintain_column_froms=True
        ).order_by(None)
        result = await self._execute_query(count_statement)
        return result.scalar_one()

    async def _execute_query(self, statement: Select) -> Result:
        return await self.session.execute(statement)

    async def _execute_statement(self, statement: Executable) -> Result:
        result = await self.session.execute(statement)
        await self.session.commit()
        return result


class UUIDRepositoryMixin(Generic[M_UUID]):
    async def get_by_id(
        self: UUIDRepositoryProtocol[M_UUID],
        id: UUID4,
        options: Sequence[Any] | None = None,
    ) -> M_UUID | None:
        statement = select(self.model).where(self.model.id == id)

        if options is not None:
            statement = statement.options(*options)

        return await self.get_one_or_none(statement)


class ExpiresAtMixin(Generic[M_EXPIRES_AT]):
    async def delete_expired(self: ExpiresAtRepositoryProtocol[M_EXPIRES_AT]):
        statement = delete(self.model).where(
            self.model.expires_at < datetime.now(timezone.utc)
        )
        await self._execute_statement(statement)


REPOSITORY = TypeVar("REPOSITORY", bound=BaseRepository)


def get_repository(
    repository_class: type[REPOSITORY], session: AsyncSession
) -> REPOSITORY:
    return repository_class(session)



import uuid
from datetime import date, datetime, timezone
from typing import TYPE_CHECKING, Any, Generic, TypeVar

from fastapi_users import schemas
from pydantic import UUID4, Field, StrictBool, StrictInt, StrictStr
from pydantic.generics import GenericModel

from fief.schemas.generics import Address, BaseModel, CreatedUpdatedAt, Timezone
from fief.schemas.tenant import TenantEmbedded

if TYPE_CHECKING:  # pragma: no cover
    from fief.models.tenant import Tenant


class UserRead(schemas.BaseUser, CreatedUpdatedAt):
    tenant_id: UUID4
    tenant: TenantEmbedded
    fields: dict[str, Any]

    class Config:
        orm_mode = True


class UserFields(BaseModel):
    def get_value(self, field: str) -> Any:
        value = getattr(self, field)
        if isinstance(value, BaseModel):
            return value.dict()
        return value


UF = TypeVar("UF", bound=UserFields)


class UserCreate(GenericModel, Generic[UF], schemas.BaseUserCreate):
    fields: UF = Field(default_factory=dict, exclude=True)  # type: ignore


class UserCreateInternal(UserCreate[UF], Generic[UF]):
    """
    Utility model so that we can hook into the logic of UserManager.create
    and add some attributes before persisting into database.
    """

    tenant_id: UUID4


class UserUpdate(GenericModel, Generic[UF], schemas.BaseUserUpdate):
    fields: UF | None = Field(exclude=True)


class UserEmailContext(schemas.BaseUser, CreatedUpdatedAt):
    tenant_id: UUID4
    fields: dict[
        str, Address | Timezone | StrictBool | StrictInt | StrictStr | datetime | date
    ]

    class Config:
        orm_mode = True

    @classmethod
    def create_sample(cls, tenant: "Tenant") -> "UserEmailContext":
        return cls(
            id=uuid.uuid4(),
            email="anne@bretagne.duchy",
            created_at=datetime.now(timezone.utc),
            updated_at=datetime.now(timezone.utc),
            tenant_id=tenant.id,
            fields={
                "first_name": "Anne",
                "last_name": "De Bretagne",
            },
        )



from typing import TYPE_CHECKING

from fief.db.workspace import WorkspaceEngineManager, get_workspace_session
from fief.models import Theme
from fief.repositories import ThemeRepository

if TYPE_CHECKING:  # pragma: no cover
    from fief.models import Workspace


async def init_default_theme(repository: ThemeRepository):
    default_theme = await repository.get_default()
    if default_theme is None:
        await repository.create(Theme.build_default())


async def init_workspace_default_theme(workspace: "Workspace"):
    workspace_engine_manager = WorkspaceEngineManager()
    async with get_workspace_session(workspace, workspace_engine_manager) as session:
        repository = ThemeRepository(session)
        await init_default_theme(repository)



from fastapi import APIRouter, Depends, Request, status

from fief.apps.dashboard.dependencies import (
    BaseContext,
    DatatableColumn,
    DatatableQueryParameters,
    DatatableQueryParametersGetter,
    get_base_context,
)
from fief.apps.dashboard.forms.api_key import APIKeyCreateForm
from fief.apps.dashboard.responses import HXRedirectResponse
from fief.crypto.token import generate_token
from fief.dependencies.admin_api_key import (
    get_api_key_by_id_or_404,
    get_paginated_api_keys,
)
from fief.dependencies.admin_authentication import is_authenticated_admin_session
from fief.dependencies.current_workspace import get_current_workspace
from fief.dependencies.logger import get_audit_logger
from fief.dependencies.main_repositories import get_main_repository
from fief.dependencies.pagination import PaginatedObjects
from fief.forms import FormHelper
from fief.logger import AuditLogger
from fief.models import AdminAPIKey, AuditLogMessage, Workspace
from fief.repositories import AdminAPIKeyRepository
from fief.templates import templates

router = APIRouter(dependencies=[Depends(is_authenticated_admin_session)])


async def get_columns() -> list[DatatableColumn]:
    return [
        DatatableColumn("Name", "name", "name_column", ordering="name"),
        DatatableColumn(
            "Created at", "created_at", "created_at_column", ordering="created_at"
        ),
        DatatableColumn("Actions", "actions", "actions_column"),
    ]


async def get_list_context(
    columns: list[DatatableColumn] = Depends(get_columns),
    datatable_query_parameters: DatatableQueryParameters = Depends(
        DatatableQueryParametersGetter(["name", "created_at", "actions"])
    ),
    paginated_api_keys: PaginatedObjects[AdminAPIKey] = Depends(get_paginated_api_keys),
):
    api_keys, count = paginated_api_keys
    return {
        "api_keys": api_keys,
        "count": count,
        "datatable_query_parameters": datatable_query_parameters,
        "columns": columns,
    }


@router.get("/", name="dashboard.api_keys:list")
async def list_api_keys(
    list_context=Depends(get_list_context),
    context: BaseContext = Depends(get_base_context),
):
    return templates.TemplateResponse(
        "admin/api_keys/list.html", {**context, **list_context}
    )


@router.api_route("/create", methods=["GET", "POST"], name="dashboard.api_keys:create")
async def create_api_key(
    request: Request,
    current_workspace: Workspace = Depends(get_current_workspace),
    repository: AdminAPIKeyRepository = Depends(
        get_main_repository(AdminAPIKeyRepository)
    ),
    list_context=Depends(get_list_context),
    context: BaseContext = Depends(get_base_context),
    audit_logger: AuditLogger = Depends(get_audit_logger),
):
    form_helper = FormHelper(
        APIKeyCreateForm,
        "admin/api_keys/create.html",
        request=request,
        context={**context, **list_context},
    )

    if await form_helper.is_submitted_and_valid():
        form = await form_helper.get_form()

        token, token_hash = generate_token()
        api_key = AdminAPIKey(token=token_hash, workspace_id=current_workspace.id)
        form.populate_obj(api_key)

        api_key = await repository.create(api_key)
        audit_logger.log_object_write(AuditLogMessage.OBJECT_CREATED, api_key)

        return templates.TemplateResponse(
            "admin/api_keys/token.html",
            {**context, **list_context, "api_key": api_key, "token": token},
            status_code=status.HTTP_201_CREATED,
            headers={"X-Fief-Object-Id": str(api_key.id)},
        )

    return await form_helper.get_response()


@router.api_route(
    "/{id:uuid}/delete", methods=["GET", "DELETE"], name="dashboard.api_keys:delete"
)
async def delete_api_key(
    request: Request,
    api_key: AdminAPIKey = Depends(get_api_key_by_id_or_404),
    repository: AdminAPIKeyRepository = Depends(
        get_main_repository(AdminAPIKeyRepository)
    ),
    list_context=Depends(get_list_context),
    context: BaseContext = Depends(get_base_context),
    audit_logger: AuditLogger = Depends(get_audit_logger),
):
    if request.method == "DELETE":
        await repository.delete(api_key)
        audit_logger.log_object_write(AuditLogMessage.OBJECT_DELETED, api_key)

        return HXRedirectResponse(
            request.url_for("dashboard.api_keys:list"),
            status_code=status.HTTP_204_NO_CONTENT,
        )
    else:
        return templates.TemplateResponse(
            "admin/api_keys/delete.html",
            {**context, **list_context, "api_key": api_key},
        )



import asyncio
import contextlib
from collections.abc import AsyncGenerator, Callable
from typing import AsyncContextManager, ClassVar
from urllib.parse import urlparse

import dramatiq
import jinja2
from dramatiq.brokers.redis import RedisBroker
from dramatiq.middleware import CurrentMessage
from pydantic import UUID4

from fief.db import AsyncSession
from fief.db.engine import create_async_session_maker, create_engine
from fief.db.workspace import WorkspaceEngineManager, get_workspace_session
from fief.locale import BabelMiddleware, get_babel_middleware_kwargs
from fief.logger import init_audit_logger, logger
from fief.models import Tenant, User, Workspace
from fief.models.generics import BaseModel
from fief.paths import EMAIL_TEMPLATES_DIRECTORY
from fief.repositories import (
    EmailTemplateRepository,
    TenantRepository,
    UserRepository,
    WorkspaceRepository,
)
from fief.services.email import EmailProvider
from fief.services.email_template.renderers import (
    EmailSubjectRenderer,
    EmailTemplateRenderer,
)
from fief.settings import settings

redis_parameters = urlparse(settings.redis_url)
redis_broker = RedisBroker(
    host=redis_parameters.hostname,
    port=redis_parameters.port,
    username=redis_parameters.username,
    password=redis_parameters.password,
    # Heroku Redis with TLS use self-signed certs, so we need to tinker a bit
    ssl=redis_parameters.scheme == "rediss",
    ssl_cert_reqs=None,
)
redis_broker.add_middleware(CurrentMessage())
dramatiq.set_broker(redis_broker)


SendTask = Callable[..., None]


def send_task(task: dramatiq.Actor, *args, **kwargs):
    logger.debug("Send task", task=task.actor_name)
    task.send(*args, **kwargs)


@contextlib.asynccontextmanager
async def get_main_session_task():
    main_engine = create_engine(settings.get_database_connection_parameters())
    session_maker = create_async_session_maker(main_engine)
    async with session_maker() as session:
        yield session
    await main_engine.dispose()


@contextlib.asynccontextmanager
async def get_workspace_session_task(
    workspace: Workspace,
) -> AsyncGenerator[AsyncSession, None]:
    workspace_engine_manager = WorkspaceEngineManager()
    async with get_workspace_session(workspace, workspace_engine_manager) as session:
        yield session
    await workspace_engine_manager.close_all()


email_provider = settings.get_email_provider()


class TaskError(Exception):
    pass


class ObjectDoesNotExistTaskError(TaskError):
    def __init__(self, object_type: type[BaseModel], id: str):
        super().__init__(f"{object_type.__name__} with id {id} does not exist.")


class TaskBase:
    __name__: ClassVar[str]

    def __init__(
        self,
        get_main_session: Callable[
            ..., AsyncContextManager[AsyncSession]
        ] = get_main_session_task,
        get_workspace_session: Callable[
            ..., AsyncContextManager[AsyncSession]
        ] = get_workspace_session_task,
        email_provider: EmailProvider = email_provider,
        send_task: SendTask = send_task,
    ) -> None:
        self.get_main_session = get_main_session
        self.get_workspace_session = get_workspace_session
        self.email_provider = email_provider
        self.send_task = send_task

        self.jinja_env = jinja2.Environment(
            loader=jinja2.FileSystemLoader(EMAIL_TEMPLATES_DIRECTORY), autoescape=True
        )
        self.jinja_env.add_extension("jinja2.ext.i18n")

    def __call__(self, *args, **kwargs):
        with asyncio.Runner() as runner:
            init_audit_logger(
                self.get_main_session, self.get_workspace_session, runner.get_loop()
            )
            BabelMiddleware(app=None, **get_babel_middleware_kwargs())
            logger.info("Start task", task=self.__name__)
            result = runner.run(self.run(*args, **kwargs))
            logger.info("Done task", task=self.__name__)
            return result

    async def _get_workspace(self, workspace_id: UUID4) -> Workspace:
        async with self.get_main_session() as session:
            repository = WorkspaceRepository(session)
            workspace = await repository.get_by_id(workspace_id)
            if workspace is None:
                raise TaskError()
            return workspace

    async def _get_user(self, user_id: UUID4, workspace: Workspace) -> User:
        async with self.get_workspace_session(workspace) as session:
            repository = UserRepository(session)
            user = await repository.get_by_id(user_id)
            if user is None:
                raise TaskError()
            return user

    async def _get_tenant(self, tenant_id: UUID4, workspace: Workspace) -> Tenant:
        async with self.get_workspace_session(workspace) as session:
            repository = TenantRepository(session)
            tenant = await repository.get_by_id(tenant_id)
            if tenant is None:
                raise TaskError()
            return tenant

    @contextlib.asynccontextmanager
    async def _get_email_template_renderer(
        self, workspace: Workspace
    ) -> AsyncGenerator[EmailTemplateRenderer, None]:
        async with self.get_workspace_session(workspace) as session:
            repository = EmailTemplateRepository(session)
            yield EmailTemplateRenderer(repository)

    @contextlib.asynccontextmanager
    async def _get_email_subject_renderer(
        self, workspace: Workspace
    ) -> AsyncGenerator[EmailSubjectRenderer, None]:
        async with self.get_workspace_session(workspace) as session:
            repository = EmailTemplateRepository(session)
            yield EmailSubjectRenderer(repository)



import secrets

from fastapi import APIRouter, Depends, Header, Request, status

from fief import schemas
from fief.apps.dashboard.dependencies import (
    BaseContext,
    DatatableColumn,
    DatatableQueryParameters,
    DatatableQueryParametersGetter,
    get_base_context,
)
from fief.apps.dashboard.forms.client import (
    ClientCreateForm,
    ClientLifetimesForm,
    ClientUpdateForm,
)
from fief.apps.dashboard.responses import HXRedirectResponse
from fief.crypto.jwk import generate_jwk
from fief.dependencies.admin_authentication import is_authenticated_admin_session
from fief.dependencies.client import get_client_by_id_or_404, get_paginated_clients
from fief.dependencies.logger import get_audit_logger
from fief.dependencies.pagination import PaginatedObjects
from fief.dependencies.webhooks import TriggerWebhooks, get_trigger_webhooks
from fief.dependencies.workspace_repositories import get_workspace_repository
from fief.forms import FormHelper
from fief.logger import AuditLogger
from fief.models import AuditLogMessage, Client
from fief.repositories import ClientRepository, TenantRepository
from fief.services.webhooks.models import ClientCreated, ClientDeleted, ClientUpdated
from fief.templates import templates

router = APIRouter(dependencies=[Depends(is_authenticated_admin_session)])


async def get_columns() -> list[DatatableColumn]:
    return [
        DatatableColumn("Name", "name", "name_column", ordering="name"),
        DatatableColumn("Type", "type", "type_column", ordering="client_type"),
        DatatableColumn("Tenant", "tenant", "tenant_column", ordering="tenant.name"),
        DatatableColumn(
            "Client ID", "client_id", "client_id_column", ordering="client_id"
        ),
    ]


async def get_list_context(
    columns: list[DatatableColumn] = Depends(get_columns),
    datatable_query_parameters: DatatableQueryParameters = Depends(
        DatatableQueryParametersGetter(["name", "type", "tenant", "client_id"])
    ),
    paginated_clients: PaginatedObjects[Client] = Depends(get_paginated_clients),
):
    clients, count = paginated_clients
    return {
        "clients": clients,
        "count": count,
        "datatable_query_parameters": datatable_query_parameters,
        "columns": columns,
    }


async def get_list_template(hx_combobox: bool = Header(False)) -> str:
    if hx_combobox:
        return "admin/clients/list_combobox.html"
    return "admin/clients/list.html"


@router.get("/", name="dashboard.clients:list")
async def list_clients(
    template: str = Depends(get_list_template),
    list_context=Depends(get_list_context),
    context: BaseContext = Depends(get_base_context),
):
    return templates.TemplateResponse(template, {**context, **list_context})


@router.get("/{id:uuid}", name="dashboard.clients:get")
async def get_client(
    client: Client = Depends(get_client_by_id_or_404),
    list_context=Depends(get_list_context),
    context: BaseContext = Depends(get_base_context),
):
    return templates.TemplateResponse(
        "admin/clients/get/general.html",
        {**context, **list_context, "client": client, "tab": "general"},
    )


@router.api_route(
    "/{id:uuid}/lifetimes", methods=["GET", "POST"], name="dashboard.clients:lifetimes"
)
async def client_lifetimes(
    request: Request,
    repository: ClientRepository = Depends(get_workspace_repository(ClientRepository)),
    client: Client = Depends(get_client_by_id_or_404),
    list_context=Depends(get_list_context),
    context: BaseContext = Depends(get_base_context),
    audit_logger: AuditLogger = Depends(get_audit_logger),
    trigger_webhooks: TriggerWebhooks = Depends(get_trigger_webhooks),
):
    form_helper = FormHelper(
        ClientLifetimesForm,
        "admin/clients/get/lifetimes.html",
        object=client,
        request=request,
        context={**context, **list_context, "client": client, "tab": "lifetimes"},
    )

    if await form_helper.is_submitted_and_valid():
        form = await form_helper.get_form()
        form.populate_obj(client)

        await repository.update(client)
        audit_logger.log_object_write(AuditLogMessage.OBJECT_UPDATED, client)
        trigger_webhooks(ClientUpdated, client, schemas.client.Client)

        return HXRedirectResponse(
            request.url_for("dashboard.clients:lifetimes", id=client.id)
        )

    return await form_helper.get_response()


@router.api_route("/create", methods=["GET", "POST"], name="dashboard.clients:create")
async def create_client(
    request: Request,
    repository: ClientRepository = Depends(get_workspace_repository(ClientRepository)),
    tenant_repository: TenantRepository = Depends(
        get_workspace_repository(TenantRepository)
    ),
    list_context=Depends(get_list_context),
    context: BaseContext = Depends(get_base_context),
    audit_logger: AuditLogger = Depends(get_audit_logger),
    trigger_webhooks: TriggerWebhooks = Depends(get_trigger_webhooks),
):
    form_helper = FormHelper(
        ClientCreateForm,
        "admin/clients/create.html",
        request=request,
        context={**context, **list_context},
    )

    if await form_helper.is_submitted_and_valid():
        form = await form_helper.get_form()

        tenant = await tenant_repository.get_by_id(form.data["tenant"])
        if tenant is None:
            form.tenant.errors.append("Unknown tenant.")
            return await form_helper.get_error_response(
                "Unknown tenant.", "unknown_tenant"
            )
        form.tenant.data = tenant

        client = Client()
        form.populate_obj(client)
        client = await repository.create(client)
        audit_logger.log_object_write(AuditLogMessage.OBJECT_CREATED, client)
        trigger_webhooks(ClientCreated, client, schemas.client.Client)

        return HXRedirectResponse(
            request.url_for("dashboard.clients:get", id=client.id),
            status_code=status.HTTP_201_CREATED,
            headers={"X-Fief-Object-Id": str(client.id)},
        )

    return await form_helper.get_response()


@router.api_route(
    "/{id:uuid}/edit",
    methods=["GET", "POST"],
    name="dashboard.clients:update",
)
async def update_client(
    request: Request,
    client: Client = Depends(get_client_by_id_or_404),
    repository: ClientRepository = Depends(get_workspace_repository(ClientRepository)),
    list_context=Depends(get_list_context),
    context: BaseContext = Depends(get_base_context),
    audit_logger: AuditLogger = Depends(get_audit_logger),
    trigger_webhooks: TriggerWebhooks = Depends(get_trigger_webhooks),
):
    form_helper = FormHelper(
        ClientUpdateForm,
        "admin/clients/edit.html",
        object=client,
        request=request,
        context={**context, **list_context, "client": client},
    )

    if await form_helper.is_submitted_and_valid():
        form = await form_helper.get_form()
        form.populate_obj(client)

        await repository.update(client)
        audit_logger.log_object_write(AuditLogMessage.OBJECT_UPDATED, client)
        trigger_webhooks(ClientUpdated, client, schemas.client.Client)

        return HXRedirectResponse(
            request.url_for("dashboard.clients:get", id=client.id)
        )

    return await form_helper.get_response()


@router.post(
    "/{id:uuid}/encryption-key",
    name="dashboard.clients:encryption_key",
)
async def create_encryption_key(
    client: Client = Depends(get_client_by_id_or_404),
    repository: ClientRepository = Depends(get_workspace_repository(ClientRepository)),
    context: BaseContext = Depends(get_base_context),
    audit_logger: AuditLogger = Depends(get_audit_logger),
    trigger_webhooks: TriggerWebhooks = Depends(get_trigger_webhooks),
):
    key = generate_jwk(secrets.token_urlsafe(), "enc")
    client.encrypt_jwk = key.export_public()
    await repository.update(client)
    audit_logger.log_object_write(AuditLogMessage.OBJECT_UPDATED, client)
    trigger_webhooks(ClientUpdated, client, schemas.client.Client)

    return templates.TemplateResponse(
        "admin/clients/encryption_key.html",
        {**context, "client": client, "key": key.export(as_dict=True)},
    )


@router.api_route(
    "/{id:uuid}/delete",
    methods=["GET", "DELETE"],
    name="dashboard.clients:delete",
)
async def delete_client(
    request: Request,
    client: Client = Depends(get_client_by_id_or_404),
    repository: ClientRepository = Depends(get_workspace_repository(ClientRepository)),
    list_context=Depends(get_list_context),
    context: BaseContext = Depends(get_base_context),
    audit_logger: AuditLogger = Depends(get_audit_logger),
    trigger_webhooks: TriggerWebhooks = Depends(get_trigger_webhooks),
):
    if request.method == "DELETE":
        await repository.delete(client)
        audit_logger.log_object_write(AuditLogMessage.OBJECT_DELETED, client)
        trigger_webhooks(ClientDeleted, client, schemas.client.Client)

        return HXRedirectResponse(
            request.url_for("dashboard.clients:list"),
            status_code=status.HTTP_204_NO_CONTENT,
        )
    else:
        return templates.TemplateResponse(
            "admin/clients/delete.html",
            {**context, **list_context, "client": client},
        )



from collections.abc import Callable, Coroutine
from typing import Any

from fastapi import Depends, HTTPException, Query, status
from pydantic import UUID4
from sqlalchemy import select

from fief.dependencies.pagination import (
    GetPaginatedObjects,
    Ordering,
    OrderingGetter,
    PaginatedObjects,
    Pagination,
    get_paginated_objects_getter,
    get_pagination,
)
from fief.dependencies.workspace_repositories import get_workspace_repository
from fief.models import Permission, User
from fief.repositories import PermissionRepository


async def get_paginated_permissions(
    query: str | None = Query(None),
    pagination: Pagination = Depends(get_pagination),
    ordering: Ordering = Depends(OrderingGetter()),
    repository: PermissionRepository = Depends(
        get_workspace_repository(PermissionRepository)
    ),
    get_paginated_objects: GetPaginatedObjects[Permission] = Depends(
        get_paginated_objects_getter
    ),
) -> PaginatedObjects[Permission]:
    statement = select(Permission)

    if query is not None:
        statement = statement.where(
            Permission.name.ilike(f"%{query}%")
            | Permission.codename.ilike(f"%{query}%")
        )

    return await get_paginated_objects(statement, pagination, ordering, repository)


async def get_permission_by_id_or_404(
    id: UUID4,
    repository: PermissionRepository = Depends(
        get_workspace_repository(PermissionRepository)
    ),
) -> Permission:
    permission = await repository.get_by_id(id)

    if permission is None:
        raise HTTPException(status_code=status.HTTP_404_NOT_FOUND)

    return permission


UserPermissionsGetter = Callable[[User], Coroutine[Any, Any, list[str]]]


async def get_user_permissions_getter(
    repository: PermissionRepository = Depends(
        get_workspace_repository(PermissionRepository)
    ),
) -> UserPermissionsGetter:
    async def _get_user_permissions(user: User) -> list[str]:
        permissions = await repository.list(
            repository.get_user_permissions_statement(user.id)
        )
        return [permission.codename for permission in permissions]

    return _get_user_permissions



from fastapi import Depends, HTTPException, Query, status
from pydantic import UUID4
from sqlalchemy import select

from fief.dependencies.pagination import (
    GetPaginatedObjects,
    Ordering,
    OrderingGetter,
    PaginatedObjects,
    Pagination,
    get_paginated_objects_getter,
    get_pagination,
)
from fief.dependencies.workspace_repositories import get_workspace_repository
from fief.models import Role
from fief.repositories import RoleRepository


async def get_paginated_roles(
    query: str | None = Query(None),
    pagination: Pagination = Depends(get_pagination),
    ordering: Ordering = Depends(OrderingGetter()),
    repository: RoleRepository = Depends(get_workspace_repository(RoleRepository)),
    get_paginated_objects: GetPaginatedObjects[Role] = Depends(
        get_paginated_objects_getter
    ),
) -> PaginatedObjects[Role]:
    statement = select(Role)

    if query is not None:
        statement = statement.where(Role.name.ilike(f"%{query}%"))

    return await get_paginated_objects(statement, pagination, ordering, repository)


async def get_role_by_id_or_404(
    id: UUID4,
    repository: RoleRepository = Depends(get_workspace_repository(RoleRepository)),
) -> Role:
    role = await repository.get_by_id(id)

    if role is None:
        raise HTTPException(status_code=status.HTTP_404_NOT_FOUND)

    return role
